import warnings
import torch
from omegaconf import OmegaConf
import torch.nn.functional as F
import os 
from easy_local_features.submodules.git_reasoningaccv.utils import resize_long_edge
from easy_local_features.utils import ops
import time

def sum_dicts(d1, d2):
    set_keys = set(d1.keys()).intersection(set(d2.keys()))
    return {k: d1[k] + d2[k] for k in set_keys}

def load_reasoning_from_checkpoint(checkpoint_path, weights_path=None, load_extractor=False, load_dino=False):   
    config = OmegaConf.load(checkpoint_path + "/model_config.yaml")
    if weights_path is None:
        # get the last checkpoint by time
        all_files = os.listdir(checkpoint_path)
        all_files = [f for f in all_files if f.endswith(".pt")]
        all_files = sorted(all_files, key=lambda x: os.path.getmtime(checkpoint_path + "/" + x))
        weights_path = all_files[-1]
        print(f"Loading weights from {weights_path}")
        weights_path = checkpoint_path + "/" + weights_path
    else:
        weights_path = checkpoint_path + "/" + weights_path
        
    model = ReasoningBase(config)
    weights = torch.load(weights_path, map_location="cpu")
    # remove "dino." and "extractor." and "semantic_reconstruction." from the dict 
    if 'state_dict' in weights:
        weights = weights['state_dict']

    if load_extractor:
        weights = {k: v for k, v in weights.items() if k.startswith("extractor.")}
        return weights

    if load_dino:
        weights = {k: v for k, v in weights.items() if k.startswith("dino.")}
        return weights
        
    weights = {k: v for k, v in weights.items() if not k.startswith("dino.")}
    weights = {k: v for k, v in weights.items() if not k.startswith("extractor.")}
    weights = {k: v for k, v in weights.items() if not k.startswith("semantic_reconstruction.")}
    
    model.load_state_dict(weights)
    model.eval()
    return {
        'model': model,
        'config': config,
        'weights_path': weights_path,
    }

def crop_patches(tensor, coords, size = 7):
    '''
        Crop [size x size] patches around 2D coordinates from a tensor.
    '''
    B, C, H, W = tensor.shape
    coords = coords.long()

    x, y = coords[:, 0], coords[:, 1]
    y = y.view(-1, 1, 1)
    x = x.view(-1, 1, 1)
    halfsize = size // 2
    # Create meshgrid for indexing
    x_offset, y_offset = torch.meshgrid(torch.arange(-halfsize, halfsize+1), torch.arange(-halfsize, halfsize+1), indexing='xy')
    y_offset = y_offset.to(tensor.device)
    x_offset = x_offset.to(tensor.device)

    # Compute indices around each coordinate
    y_indices = (y + y_offset.view(1, size, size)).squeeze(0) + halfsize
    x_indices = (x + x_offset.view(1, size, size)).squeeze(0) + halfsize

    # Handle out-of-boundary indices with padding
    tensor_padded = torch.nn.functional.pad(tensor, (halfsize, halfsize, halfsize, halfsize), mode='constant')

    # Index tensor to get patches
    patches = tensor_padded[:, :, y_indices, x_indices] # [B, C, N, H, W]
    return patches
 
class Transformer(torch.nn.Module):
    def __init__(self, conf):
        super().__init__()
        self.conf = conf
        self.n_heads = conf.num_heads
        
        self.attention = torch.nn.MultiheadAttention(
            embed_dim=conf.features_dim,
            num_heads=conf.num_heads,
            batch_first=True,
        )
        
        self.fc = torch.nn.Linear(conf.features_dim, conf.features_dim)
        
        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(conf.features_dim * 2, conf.features_dim * 2),
            torch.nn.ReLU(),
            torch.nn.Linear(conf.features_dim * 2, conf.features_dim * 2),
            torch.nn.ReLU(),
            torch.nn.Linear(conf.features_dim * 2, conf.features_dim),
        )
        
    def forward(self, q, k, v):
        attn_output, attn_output_weights = self.attention(q, k, v)
        attn_output = self.fc(attn_output)
        v = v + self.mlp(torch.cat([attn_output, v], dim=-1))
        return v, attn_output_weights
        

def sample_features_dino(keypoints, features, s=14, mode="bicubic"):
    b, c, h, w = features.shape
    keypoints = keypoints / (keypoints.new_tensor([w, h]) * s)
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
    features = torch.nn.functional.grid_sample(
        features, keypoints.view(b, 1, -1, 2), mode=mode, align_corners=False
    )
    features = torch.nn.functional.normalize(
        features.reshape(b, c, -1), p=2, dim=1
    )
    
    features = features.permute(0, 2, 1)
    return features
        
class ReasoningBase(torch.nn.Module):
    default_conf = {
        
        "dino":{
            "weights": "dinov2_vits14", # dinov2_vits14, dinov2_vitb14, dinov2_vitl14, dinov2_vitg14
            "allow_resize": True,
        },
        
        "extractor":{
            "model_name": "aliked-n16",
            "max_num_keypoints": 2048,
            "detection_threshold": 0.0,
            "force_num_keypoints": True,
            "pretrained": True,
            "nms_radius": 2,
        },
        
        "reasoning":{
            "extractor_features_dim": None,
            "semantic_features_dim": None,
            "features_dim": 256,
            "n_attention_layers": 14,
            "num_heads": 4, 
            "attention": "full", # full, linear
        },
        "semantic_interpolation_mode": "bicubic",
        "activate_timers": False,
        "attention_progression": "alternating", # alternating, semantic, visual
        "deep_supervision": True,
        "semantic_conditioning": False,
        "fix_dino_size": -1,
    }
    
    required_data_keys = ["image"]

    def __init__(self, conf={}):
        super().__init__()
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.dino_reprojection = torch.nn.Linear(conf.reasoning.semantic_features_dim, conf.reasoning.features_dim)
        self.extractor_reprojection = torch.nn.Linear(conf.reasoning.extractor_features_dim, conf.reasoning.features_dim)
        
        self.texture_attentions = torch.nn.ModuleList([
            Transformer(conf.reasoning)
            for _ in range(conf.reasoning.n_attention_layers)
        ])

        self.semantic_attentions = torch.nn.ModuleList([
            Transformer(conf.reasoning)
            for _ in range(conf.reasoning.n_attention_layers)
        ])
        
        self.joint_features = torch.nn.Linear(conf.reasoning.features_dim * 2, conf.reasoning.features_dim)


    @classmethod
    def from_experiment(self, checkpoint_path, weights_path=None):
        return load_reasoning_from_checkpoint(checkpoint_path, weights_path)
        
    def forward(self, data, keypoints=None):
        '''data = {
            'image': torch.Tensor,
            'keypoints': torch.Tensor,
            'descriptors': torch.Tensor,
            'semantic_features': torch.Tensor,
        }
        '''
            
        dino_features = data['semantic_features']
        extractor_descriptors = data['descriptors']
        keypoints = data['keypoints']
        
        # reproject features to common space
        proj_dino_features = self.dino_reprojection(dino_features)
        proj_extractor_descriptors = self.extractor_reprojection(extractor_descriptors)
        
        # all_reasoning_descriptors, all_layers_attn_output_weights = self.reasoning(proj_dino_features, proj_extractor_descriptors)
        all_reasoning_descriptors, all_layers_attn_output_weights = self.reasoning(proj_extractor_descriptors, proj_dino_features, self.texture_attentions, 'alternating')
        all_semantic_descriptors, all_semantic_attn_output_weights  = self.reasoning(proj_dino_features, proj_dino_features, self.semantic_attentions, 'semantic')

        return_data =  {
            # final features
            "keypoints": keypoints,
            "reasoning_features": all_reasoning_descriptors[-1],
            "semantic_features": all_semantic_descriptors[-1],
            # intermediate features
            "all_reasoning_features": all_reasoning_descriptors,
            "all_semantic_features": all_semantic_descriptors,
            # base features                
            "dino_features": dino_features,
            "extractor_features": extractor_descriptors,
        }
            
        return return_data
    
    def reasoning(self, start_feature, query_feature, attention_layers, attention_progression):
        reasoning_descriptors = start_feature.clone()
        all_attention_weights = []
        all_reasoning_descriptors = []
        if attention_progression == "alternating":
            for layer_idx in range(len(attention_layers)):
                if layer_idx % 2 == 0:
                    reasoning_descriptors, attn_output_weights = attention_layers[layer_idx](
                        q=query_feature, 
                        k=reasoning_descriptors, 
                        v=reasoning_descriptors,
                    )

                else:
                    reasoning_descriptors, attn_output_weights = attention_layers[layer_idx](
                        q=start_feature, 
                        k=reasoning_descriptors, 
                        v=reasoning_descriptors,
                    )
                
                all_attention_weights.append(attn_output_weights.clone())
                all_reasoning_descriptors.append(reasoning_descriptors.clone())
        elif attention_progression == "semantic":
            for layer_idx in range(len(attention_layers)):
                reasoning_descriptors, attn_output_weights = attention_layers[layer_idx](
                    q=query_feature,
                    k=reasoning_descriptors, 
                    v=reasoning_descriptors,
                )
                
                all_attention_weights.append(attn_output_weights.clone())
                all_reasoning_descriptors.append(reasoning_descriptors.clone())
        else:                
            raise ValueError(f"Unknown attention progression {self.conf.attention_progression}")

        all_reasoning_descriptors = torch.stack(all_reasoning_descriptors, dim=0)
        return all_reasoning_descriptors, all_attention_weights

    def loss(self, data, pred, args):
        loss = 0
        log_dict = {}

        softmax_loss = self.dual_softmax_loss(pred, data)
        log_dict['0_train/loss_softmax'] =  softmax_loss
        loss += softmax_loss['matchable'] * 0.25
        loss += softmax_loss['unmatchable'] * 0.1
            
        return loss, log_dict

    def match(self, data, pred=None):
        @torch.no_grad()
        def find_nn(sim, ratio_thresh, distance_thresh):
            sim_nn, ind_nn = sim.topk(2 if ratio_thresh else 1, dim=-1, largest=True)
            dist_nn = 2 * (1 - sim_nn)
            mask = torch.ones(ind_nn.shape[:-1], dtype=torch.bool, device=sim.device)
            if ratio_thresh:
                mask = mask & (dist_nn[..., 0] <= (ratio_thresh**2) * dist_nn[..., 1])
            if distance_thresh:
                mask = mask & (dist_nn[..., 0] <= distance_thresh**2)
            matches = torch.where(mask, ind_nn[..., 0], ind_nn.new_tensor(-1))
            return matches

        def mutual_check(m0, m1):
            inds0 = torch.arange(m0.shape[-1], device=m0.device)
            inds1 = torch.arange(m1.shape[-1], device=m1.device)
            loop0 = torch.gather(m1, -1, torch.where(m0 > -1, m0, m0.new_tensor(0)))
            loop1 = torch.gather(m0, -1, torch.where(m1 > -1, m1, m1.new_tensor(0)))
            m0_new = torch.where((m0 > -1) & (inds0 == loop0), m0, m0.new_tensor(-1))
            m1_new = torch.where((m1 > -1) & (inds1 == loop1), m1, m1.new_tensor(-1))
            return m0_new, m1_new
        
        
        if pred is None:
            with torch.inference_mode():        
                pred0 = self({"image": data['image0']})
                pred1 = self({"image": data['image1']})

            pred = {}
            pred.update({k+"0": v for k, v in pred0.items()})
            pred.update({k+"1": v for k, v in pred1.items()})
            
        f0 = pred['reasoning_features0']
        f1 = pred['reasoning_features1']
        s0 = pred['semantic_features0']
        s1 = pred['semantic_features1']
        
        # normalize the features
        f0 = F.normalize(f0, dim=-1)
        f1 = F.normalize(f1, dim=-1)
        s0 = F.normalize(s0, dim=-1)
        s1 = F.normalize(s1, dim=-1)
        
        reasoning_sim = torch.einsum("bnd,bmd->bnm", f0, f1)
        semantic_sim = torch.einsum("bnd,bmd->bnm", s0, s1)
        if self.conf.semantic_conditioning:
            sim = reasoning_sim * semantic_sim
        else:
            sim = reasoning_sim
        
        matches0 = find_nn(sim, None, None)
        matches1 = find_nn(sim.transpose(1, 2), None, None)
        matches0, matches1 = mutual_check(matches0, matches1)
        
        return_dict = {
            'matches0': [],
            'matches1': [],
        }
        
        b_size = sim.shape[0]
        for b in range(b_size):
            keypoints0 = pred['keypoints0'][b]
            keypoints1 = pred['keypoints1'][b]      
            m0 = matches0[b]  

            valid = m0 > -1
            
            if valid.sum() == 0:
                print(f"No matches found for image {b}")
            
            mkpts0 = keypoints0[valid]
            mkpts1 = keypoints1[m0[valid]]
            
            return_dict['matches0'].append(mkpts0)
            return_dict['matches1'].append(mkpts1)
            
        return return_dict

class Reasoning(torch.nn.Module):
    def __init__(self, reasoning_model, dev=None):
        super().__init__()
        self.conf = conf = reasoning_model.conf
        if dev is not None:
            dev = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            
        self.dev = dev
        print(f'Looking for "{conf.extractor.model_name}"')

        with warnings.catch_warnings():
            from easy_local_features.submodules.git_reasoningaccv.dinov2 import DinoV2
            self.dino = DinoV2(conf.dino).eval()
        if 'aliked-n' in conf.extractor.model_name:
            from easy_local_features.submodules.git_reasoningaccv.aliked import ALIKED
            assert conf.extractor.model_name in ALIKED.cfgs, f"Model {conf.extractor.model_name} not found in ALIKED.cfgs"
            self.extractor = ALIKED(conf.extractor).eval()     
        elif 'alike-n' in conf.extractor.model_name:
            from easy_local_features.feature.baseline_alike import ALIKE_baseline
            self.extractor = ALIKE_baseline(
                {
                    "model_name":'alike-n',
                    "top_k":self.conf.extractor.max_num_keypoints,
                    "n_limit":self.conf.extractor.max_num_keypoints,
                }
            )
            self.extractor.to(self.dev)
        elif 'dedode' in conf.extractor.model_name:
            from kornia.feature import DeDoDe
            assert conf.extractor.model_name in ["dedode-G", "dedode-B"], f"Model {conf.extractor.model_name} not found in DeDoDe"
            descriptor_type = conf.extractor.model_name.split("-")[1]
            self.extractor = DeDoDe.from_pretrained(detector_weights="L-upright", descriptor_weights=f"{descriptor_type}-upright")
        elif 'superpoint' in conf.extractor.model_name:
            from easy_local_features.submodules.git_reasoningaccv.superpoint import SuperPoint
            self.extractor = SuperPoint(conf.extractor).eval()
                
        elif 'xfeat' in conf.extractor.model_name:
            self.extractor = torch.hub.load('verlab/accelerated_features', 'XFeat', pretrained = True, top_k = conf.extractor.max_num_keypoints, detection_threshold=0.)
       
        elif 'relf' in conf.extractor.model_name:
            from easy_local_features.feature.baseline_superpoint import SuperPoint_baseline
            from easy_local_features.feature.baseline_relf import RELF_baseline
            self.detector = SuperPoint_baseline({
                "top_k":conf.extractor.max_num_keypoints,
                "legacy_sampling": False,
            })
            self.extractor = RELF_baseline({
                'model': 're_resnet'
            })
            self.extractor.to(self.dev)
            self.detector.to(self.dev)
        else:
            raise ValueError(f"Model {conf.extractor.model_name} not found")
        
        self.reasoning_model = reasoning_model
        
    def forward(self, data):
        start_extract = time.time()
        if 'xfeat' in self.conf.extractor.model_name:
            response = self.extractor.detectAndCompute(data['image'], top_k = self.conf.extractor.max_num_keypoints)
            extractor_pred = {
                'keypoints': response[0]['keypoints'].unsqueeze(0),
                'descriptors': response[0]['descriptors'].unsqueeze(0),
            }
        elif 'dedode' in self.conf.extractor.model_name:
            keypoints, scores, features = self.extractor(data['image'], n=self.conf.extractor.max_num_keypoints, apply_imagenet_normalization=True)
            extractor_pred = {"keypoints": keypoints, "descriptors": features}                        
        elif self.conf.extractor.model_name in ['superpoint', 'aliked-n']:
            response = self.extractor(data)
            extractor_pred = response
        elif 'alike-n' in self.conf.extractor.model_name:
            keypoints, descriptors = self.extractor.detectAndCompute(data['image'])
            extractor_pred = {
                'keypoints': keypoints.unsqueeze(0),
                'descriptors': descriptors.unsqueeze(0),
            }
        elif 'relf' in self.conf.extractor.model_name:
            images = data['image'].to(self.dev)
            keypoints = self.detector.detect(images)
            keypoints, descriptors = self.extractor.compute(images, keypoints)
            extractor_pred = {
                'keypoints': keypoints,
                'descriptors': descriptors,
            }
        else: 
            raise ValueError(f"Model {self.conf.extractor.model_name} not found")
        end_extract = time.time()
        
        start_dino = time.time()
        resized, scale = resize_long_edge(data['image'], 896)
        if self.conf.fix_dino_size > 0:
            resized, scale = resize_long_edge(data['image'], self.conf.fix_dino_size)

        semantic_features = self.dino({'image': resized})
        dino_features = sample_features_dino(extractor_pred['keypoints']*scale, semantic_features['features'], mode=self.conf.semantic_interpolation_mode)
        end_dino = time.time()
        
        start_reasoning = time.time()        
        reasoning_features = self.reasoning_model({
            'image': data['image'],
            'keypoints': extractor_pred['keypoints'],
            'descriptors': extractor_pred['descriptors'],
            'semantic_features': dino_features,
        }) 
        end_reasoning = time.time()
        
        time_dict = {
            'extract': end_extract - start_extract,
            'dino': end_dino - start_dino,
            'reasoning': end_reasoning - start_reasoning,
        }
        reasoning_features['time'] = time_dict
        reasoning_features['dino_map'] = semantic_features['features']
        
        return reasoning_features
    
    @torch.inference_mode()
    def detectAndCompute(self, image):
        image = ops.prepareImage(image).to(self.dev)
        response = self.forward({'image': image})
        
        keypoints = response['keypoints'] # [B, N, 2]
        descritexture_features = response['reasoning_features'] # [B, N, D]
        semantic_features = response['semantic_features'] # [B, N, D]
        descriptors = torch.stack([descritexture_features, semantic_features], dim=-1) # [B, N, D, 2]
        
        return keypoints, descriptors
        
    @torch.inference_mode()
    def match(self, data, pred={}):
        def find_nn(sim, ratio_thresh, distance_thresh):
            sim_nn, ind_nn = sim.topk(2 if ratio_thresh else 1, dim=-1, largest=True)
            dist_nn = 2 * (1 - sim_nn)
            mask = torch.ones(ind_nn.shape[:-1], dtype=torch.bool, device=sim.device)
            if ratio_thresh:
                mask = mask & (dist_nn[..., 0] <= (ratio_thresh**2) * dist_nn[..., 1])
            if distance_thresh:
                mask = mask & (dist_nn[..., 0] <= distance_thresh**2)
            matches = torch.where(mask, ind_nn[..., 0], ind_nn.new_tensor(-1))
            return matches

        def mutual_check(m0, m1):
            inds0 = torch.arange(m0.shape[-1], device=m0.device)
            inds1 = torch.arange(m1.shape[-1], device=m1.device)
            loop0 = torch.gather(m1, -1, torch.where(m0 > -1, m0, m0.new_tensor(0)))
            loop1 = torch.gather(m0, -1, torch.where(m1 > -1, m1, m1.new_tensor(0)))
            m0_new = torch.where((m0 > -1) & (inds0 == loop0), m0, m0.new_tensor(-1))
            m1_new = torch.where((m1 > -1) & (inds1 == loop1), m1, m1.new_tensor(-1))
            return m0_new, m1_new

        if pred == {}:
            pred0 = self({
                "image": data['image0']
            })
            pred1 = self({
                "image": data['image1']
            })

            pred = {}
            pred.update({k+"0": v for k, v in pred0.items()})
            pred.update({k+"1": v for k, v in pred1.items()})
            
        f0 = pred['reasoning_features0']
        f1 = pred['reasoning_features1']
        s0 = pred['semantic_features0']
        s1 = pred['semantic_features1']
        
        # # normalize the features
        f0 = F.normalize(f0, dim=-1)
        f1 = F.normalize(f1, dim=-1)
        s0 = F.normalize(s0, dim=-1)
        s1 = F.normalize(s1, dim=-1)
        
        reasoning_sim = torch.einsum("bnd,bmd->bnm", f0, f1)
        semantic_sim = torch.einsum("bnd,bmd->bnm", s0, s1)

        if self.conf.semantic_conditioning:
            sim = reasoning_sim * semantic_sim
        else:
            sim = reasoning_sim
        
        matches0 = find_nn(sim, None, None)
        matches1 = find_nn(sim.transpose(1, 2), None, None)
        matches0, matches1 = mutual_check(matches0, matches1)
        
        return_dict = {
            'matches0': [],
            'matches1': [],
            'pred': pred,
        }
        
        b_size = sim.shape[0]
        for b in range(b_size):
            keypoints0 = pred['keypoints0'][b]
            keypoints1 = pred['keypoints1'][b]      
            m0 = matches0[b]  
                
            valid = m0 > -1
            
            if valid.sum() == 0:
                print(f"No matches found for image {b}")
            
            mkpts0 = keypoints0[valid]
            mkpts1 = keypoints1[m0[valid]]
            
            return_dict['matches0'].append(mkpts0)
            return_dict['matches1'].append(mkpts1)
            
        return return_dict
    


# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
from PIL import Image
import numpy as np

import torch
import torchvision.transforms as tvf

from .transforms import instanciate_transformation
from .transforms_tools import persp_apply

RGB_mean = [0.485, 0.456, 0.406]
RGB_std  = [0.229, 0.224, 0.225]

norm_RGB = tvf.Compose([tvf.ToTensor(), tvf.Normalize(mean=RGB_mean, std=RGB_std)])


class PairLoader:
    """ On-the-fly jittering of pairs of image with dense pixel ground-truth correspondences.
    
    crop:   random crop applied to both images
    scale:  random scaling applied to img2
    distort: random ditorsion applied to img2
    
    self[idx] returns a dictionary with keys: img1, img2, aflow, mask
     - img1: cropped original
     - img2: distorted cropped original
     - aflow: 'absolute' optical flow = (x,y) position of each pixel from img1 in img2
     - mask: (binary image) valid pixels of img1
    """
    def __init__(self, dataset, crop='', scale='', distort='', norm = norm_RGB, 
                       what = 'aflow mask', idx_as_rng_seed = False):
        assert hasattr(dataset, 'npairs')
        assert hasattr(dataset, 'get_pair')
        self.dataset = dataset
        self.distort = instanciate_transformation(distort)
        self.crop = instanciate_transformation(crop)
        self.norm = instanciate_transformation(norm)
        self.scale = instanciate_transformation(scale)
        self.idx_as_rng_seed = idx_as_rng_seed # to remove randomness
        self.what = what.split() if isinstance(what, str) else what
        self.n_samples = 5 # number of random trials per image

    def __len__(self):
        assert len(self.dataset) == self.dataset.npairs, pdb.set_trace() # and not nimg
        return len(self.dataset)

    def __repr__(self):
        fmt_str = 'PairLoader\n'
        fmt_str += repr(self.dataset)
        fmt_str += '  npairs: %d\n' % self.dataset.npairs
        short_repr = lambda s: repr(s).strip().replace('\n',', ')[14:-1].replace('    ',' ')
        fmt_str += '  Distort: %s\n' % short_repr(self.distort)
        fmt_str += '  Crop: %s\n' % short_repr(self.crop)
        fmt_str += '  Norm: %s\n' % short_repr(self.norm)
        return fmt_str

    def __getitem__(self, i):
        #from time import time as now; t0 = now()
        if self.idx_as_rng_seed:
            import random
            random.seed(i)
            np.random.seed(i)

        # Retrieve an image pair and their absolute flow
        img_a, img_b, metadata = self.dataset.get_pair(i, self.what)
        
        # aflow contains pixel coordinates indicating where each 
        # pixel from the left image ended up in the right image
        # as (x,y) pairs, but its shape is (H,W,2)
        aflow = np.float32(metadata['aflow'])
        mask = metadata.get('mask', np.ones(aflow.shape[:2],np.uint8))

        # apply transformations to the second image
        img_b = {'img': img_b, 'persp':(1,0,0,0,1,0,0,0)}
        if self.scale:
            img_b = self.scale(img_b)
        if self.distort:
            img_b = self.distort(img_b)
        
        # apply the same transformation to the flow
        aflow[:] = persp_apply(img_b['persp'], aflow.reshape(-1,2)).reshape(aflow.shape)
        corres = None
        if 'corres' in metadata:
            corres = np.float32(metadata['corres'])
            corres[:,1] = persp_apply(img_b['persp'], corres[:,1])
        
        # apply the same transformation to the homography
        homography = None
        if 'homography' in metadata:
            homography = np.float32(metadata['homography'])
            # p_b = homography * p_a
            persp = np.float32(img_b['persp']+(1,)).reshape(3,3)
            homography = persp @ homography

        # determine crop size
        img_b = img_b['img']
        crop_size = self.crop({'imsize':(10000,10000)})['imsize']
        output_size_a = min(img_a.size, crop_size)
        output_size_b = min(img_b.size, crop_size)
        img_a = np.array(img_a)
        img_b = np.array(img_b)

        ah,aw,p1 = img_a.shape
        bh,bw,p2 = img_b.shape
        assert p1 == 3
        assert p2 == 3
        assert aflow.shape == (ah, aw, 2)
        assert mask.shape == (ah, aw)

        # Let's start by computing the scale of the
        # optical flow and applying a median filter:
        dx = np.gradient(aflow[:,:,0])
        dy = np.gradient(aflow[:,:,1])
        scale = np.sqrt(np.clip(np.abs(dx[1]*dy[0] - dx[0]*dy[1]), 1e-16, 1e16))

        accu2 = np.zeros((16,16), bool)
        Q = lambda x, w: np.int32(16 * (x - w.start) / (w.stop - w.start))
        
        def window1(x, size, w):
            l = x - int(0.5 + size / 2)
            r = l + int(0.5 + size)
            if l < 0: l,r = (0, r - l)
            if r > w: l,r = (l + w - r, w)
            if l < 0: l,r = 0,w # larger than width
            return slice(l,r)
        def window(cx, cy, win_size, scale, img_shape):
            return (window1(cy, win_size[1]*scale, img_shape[0]), 
                    window1(cx, win_size[0]*scale, img_shape[1]))

        n_valid_pixel = mask.sum()
        sample_w = mask / (1e-16 + n_valid_pixel)
        def sample_valid_pixel():
            n = np.random.choice(sample_w.size, p=sample_w.ravel())
            y, x = np.unravel_index(n, sample_w.shape)
            return x, y
        
        # Find suitable left and right windows
        trials = 0 # take the best out of few trials
        best = -np.inf, None
        for _ in range(50*self.n_samples):
            if trials >= self.n_samples: break # finished!

            # pick a random valid point from the first image
            if n_valid_pixel == 0: break
            c1x, c1y = sample_valid_pixel()
            
            # Find in which position the center of the left
            # window ended up being placed in the right image
            c2x, c2y = (aflow[c1y, c1x] + 0.5).astype(np.int32)
            if not(0 <= c2x < bw and 0 <= c2y < bh): continue

            # Get the flow scale
            sigma = scale[c1y, c1x]

            # Determine sampling windows
            if 0.2 < sigma < 1: 
                win1 = window(c1x, c1y, output_size_a, 1/sigma, img_a.shape)
                win2 = window(c2x, c2y, output_size_b, 1, img_b.shape)
            elif 1 <= sigma < 5:
                win1 = window(c1x, c1y, output_size_a, 1, img_a.shape)
                win2 = window(c2x, c2y, output_size_b, sigma, img_b.shape)
            else:
                continue # bad scale

            # compute a score based on the flow
            x2,y2 = aflow[win1].reshape(-1, 2).T.astype(np.int32)
            # Check the proportion of valid flow vectors
            valid = (win2[1].start <= x2) & (x2 < win2[1].stop) \
                  & (win2[0].start <= y2) & (y2 < win2[0].stop)
            score1 = (valid * mask[win1].ravel()).mean()
            # check the coverage of the second window
            accu2[:] = False
            accu2[Q(y2[valid],win2[0]), Q(x2[valid],win2[1])] = True
            score2 = accu2.mean()
            # Check how many hits we got
            score = min(score1, score2)

            trials += 1
            if score > best[0]:
                best = score, win1, win2
        
        if None in best: # counldn't find a good window
            img_a = np.zeros(output_size_a[::-1]+(3,), dtype=np.uint8)
            img_b = np.zeros(output_size_b[::-1]+(3,), dtype=np.uint8)
            aflow = np.nan * np.ones((2,)+output_size_a[::-1], dtype=np.float32)
            homography = np.nan * np.ones((3,3), dtype=np.float32)

        else:
            win1, win2 = best[1:]
            img_a = img_a[win1]
            img_b = img_b[win2]
            aflow = aflow[win1] - np.float32([[[win2[1].start, win2[0].start]]])
            mask = mask[win1]
            aflow[~mask.view(bool)] = np.nan # mask bad pixels!
            aflow = aflow.transpose(2,0,1) # --> (2,H,W)
            
            if corres is not None:
                corres[:,0] -= (win1[1].start, win1[0].start)
                corres[:,1] -= (win2[1].start, win2[0].start)
            
            if homography is not None:
                trans1 = np.eye(3, dtype=np.float32)
                trans1[:2,2] = (win1[1].start, win1[0].start)
                trans2 = np.eye(3, dtype=np.float32)
                trans2[:2,2] = (-win2[1].start, -win2[0].start)
                homography = trans2 @ homography @ trans1
                homography /= homography[2,2]
            
            # rescale if necessary
            if img_a.shape[:2][::-1] != output_size_a:
                sx, sy = (np.float32(output_size_a)-1)/(np.float32(img_a.shape[:2][::-1])-1)
                img_a = np.asarray(Image.fromarray(img_a).resize(output_size_a, Image.ANTIALIAS))
                mask = np.asarray(Image.fromarray(mask).resize(output_size_a, Image.NEAREST))
                afx = Image.fromarray(aflow[0]).resize(output_size_a, Image.NEAREST)
                afy = Image.fromarray(aflow[1]).resize(output_size_a, Image.NEAREST)
                aflow = np.stack((np.float32(afx), np.float32(afy)))
                
                if corres is not None:
                    corres[:,0] *= (sx, sy)
                
                if homography is not None:
                    homography = homography @ np.diag(np.float32([1/sx,1/sy,1]))
                    homography /= homography[2,2]

            if img_b.shape[:2][::-1] != output_size_b:
                sx, sy = (np.float32(output_size_b)-1)/(np.float32(img_b.shape[:2][::-1])-1)
                img_b = np.asarray(Image.fromarray(img_b).resize(output_size_b, Image.ANTIALIAS))
                aflow *= [[[sx]], [[sy]]]
                
                if corres is not None:
                    corres[:,1] *= (sx, sy)
                
                if homography is not None:
                    homography = np.diag(np.float32([sx,sy,1])) @ homography
                    homography /= homography[2,2]
    
        assert aflow.dtype == np.float32, pdb.set_trace()
        assert homography is None or homography.dtype == np.float32, pdb.set_trace()
        if 'flow' in self.what:
            H, W = img_a.shape[:2]
            mgrid = np.mgrid[0:H, 0:W][::-1].astype(np.float32)
            flow = aflow - mgrid
        
        result = dict(img1=self.norm(img_a), img2=self.norm(img_b))
        for what in self.what:
            try: result[what] = eval(what)
            except NameError: pass
        return result



def threaded_loader( loader, iscuda, threads, batch_size=1, shuffle=True):
    """ Get a data loader, given the dataset and some parameters.
    
    Parameters
    ----------
    loader : object[i] returns the i-th training example.
    
    iscuda : bool
        
    batch_size : int
    
    threads : int
    
    shuffle : int
    
    Returns
    -------
        a multi-threaded pytorch loader.
    """
    return torch.utils.data.DataLoader(
        loader,
        batch_size = batch_size,
        shuffle = shuffle,
        sampler = None,
        num_workers = threads,
        pin_memory = iscuda,
        collate_fn=collate)



def collate(batch, _use_shared_memory=True):
    """Puts each data field into a tensor with outer dimension batch size.
    Copied from https://github.com/pytorch in torch/utils/data/_utils/collate.py
    """
    import re
    error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
    elem_type = type(batch[0])
    if isinstance(batch[0], torch.Tensor):
        out = None
        if _use_shared_memory:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])
            storage = batch[0].storage()._new_shared(numel)
            out = batch[0].new(storage)
        return torch.stack(batch, 0, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = batch[0]
        assert elem_type.__name__ == 'ndarray'
        # array of string classes and object
        if re.search('[SaUO]', elem.dtype.str) is not None:
            raise TypeError(error_msg.format(elem.dtype))
        batch = [torch.from_numpy(b) for b in batch]
        try:
            return torch.stack(batch, 0)
        except RuntimeError:
            return batch
    elif batch[0] is None:
        return list(batch)
    elif isinstance(batch[0], int):
        return torch.LongTensor(batch)
    elif isinstance(batch[0], float):
        return torch.DoubleTensor(batch)
    elif isinstance(batch[0], str):
        return batch
    elif isinstance(batch[0], dict):
        return {key: collate([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], (tuple,list)):
        transposed = zip(*batch)
        return [collate(samples) for samples in transposed]

    raise TypeError((error_msg.format(type(batch[0]))))



def tensor2img(tensor, model=None):
    """ convert back a torch/numpy tensor to a PIL Image
        by undoing the ToTensor() and Normalize() transforms.
    """
    mean = norm_RGB.transforms[1].mean
    std =  norm_RGB.transforms[1].std
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.detach().cpu().numpy()
    
    res = np.uint8(np.clip(255*((tensor.transpose(1,2,0) * std) + mean), 0, 255))
    from PIL import Image
    return Image.fromarray(res)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser("Tool to debug/visualize the data loader")
    parser.add_argument("dataloader", type=str, help="command to create the data loader")
    args = parser.parse_args()

    from datasets import *
    auto_pairs = lambda db: SyntheticPairDataset(db,
        'RandomScale(256,1024,can_upscale=True)', 
        'RandomTilting(0.5), PixelNoise(25)')
        
    loader = eval(args.dataloader)
    print("Data loader =", loader)

    from tools.viz import show_flow
    for data in loader:
        aflow = data['aflow']
        H, W = aflow.shape[-2:]
        flow = (aflow - np.mgrid[:H, :W][::-1]).transpose(1,2,0)
        show_flow(tensor2img(data['img1']), tensor2img(data['img2']), flow)




import torch

# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

# @dimchecked
def cross_product_matrix(v):
    ''' following
        en.wikipedia.org/wiki/Cross_product#Conversion_to_matrix_multiplication
    '''

    return torch.tensor([
        [    0, -v[2],  v[1]],
        [ v[2],     0, -v[0]],
        [-v[1],  v[0],     0]
    ], dtype=v.dtype, device=v.device)

# @dimchecked
def xy_to_xyw(xy):
    ones = torch.ones(1, xy.shape[1], device=xy.device, dtype=xy.dtype)
    return torch.cat([xy, ones], dim=0)

# @dimchecked
def ims2E(im1, im2):
    R = im2.R @ im1.R.T
    T = im2.T - R @ im1.T
    return cross_product_matrix(T) @ R

# @dimchecked
def ims2F(im1, im2):
    E = ims2E(im1, im2)
    return im2.K_inv.T @ E @ im1.K_inv
 
# @dimchecked
def symdimm(x1, x2, im1, im2):
    x1n = im1.K_inv @ xy_to_xyw(x1)
    x2n = im2.K_inv @ xy_to_xyw(x2)

    E = ims2E(im1, im2)

    E_x1  = E @ x1n
    Et_x2 = E.T @ x2n
    x2_E_x1 = x2n.T @ E_x1

    n = lambda v: torch.norm(v, p=2, dim=0)

    n1 = 1 / n(E_x1[:2])[None, :]
    n2 = 1 / n(Et_x2[:2])[:, None]
    norm = n1 + n2
    dist = x2_E_x1.pow(2) * norm
    return dist.T

# @dimchecked
def asymmdist(x1, x2, F):
    '''
    following http://www.cs.toronto.edu/~jepson/csc420/notes/epiPolarGeom.pdf
    (page 12)
    '''

    x1_h = xy_to_xyw(x1)
    x2_h = xy_to_xyw(x2)

    Ft_x2 = F.T @ x2_h
    norm  = torch.norm(Ft_x2[:2], p=2, dim=0)
    dist  = (Ft_x2 / norm).T @ x1_h
    return dist.T

# @dimchecked
def asymmdist_from_imgs(x1, x2, im1, im2):
    F = ims2F(im1, im2)
    return asymmdist(x1, x2, F)

# @dimchecked
def p_asymmdist(x1, x2, F) :
    '''
    following http://www.cs.toronto.edu/~jepson/csc420/notes/epiPolarGeom.pdf
    (page 12)
    '''

    x1_h = xy_to_xyw(x1)
    x2_h = xy_to_xyw(x2)

    Ft_x2 = F.T @ x2_h
    norm  = torch.norm(Ft_x2[:2], p=2, dim=0)
    Ft_x2_n = Ft_x2 / norm

    return torch.einsum('ca,ca->a', (Ft_x2_n, x1_h))

# @dimchecked
def p_asymmdist_from_imgs(x1, x2, im1, im2) :
    F = ims2F(im1, im2)
    return p_asymmdist(x1, x2, F)



from typing import Optional, Union
import torch
from torch import device
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as tvm
import gc
from romatch.utils.utils import get_autocast_params


class ResNet50(nn.Module):
    def __init__(self, pretrained=False, high_res = False, weights = None, 
                 dilation = None, freeze_bn = True, anti_aliased = False, early_exit = False, amp = False, amp_dtype = torch.float16) -> None:
        super().__init__()
        if dilation is None:
            dilation = [False,False,False]
        if anti_aliased:
            pass
        else:
            if weights is not None:
                self.net = tvm.resnet50(weights = weights,replace_stride_with_dilation=dilation)
            else:
                self.net = tvm.resnet50(pretrained=pretrained,replace_stride_with_dilation=dilation)
            
        self.high_res = high_res
        self.freeze_bn = freeze_bn
        self.early_exit = early_exit
        self.amp = amp
        self.amp_dtype = amp_dtype

    def forward(self, x, **kwargs):
        autocast_device, autocast_enabled, autocast_dtype = get_autocast_params(x.device, self.amp, self.amp_dtype)
        with torch.autocast(autocast_device, enabled=autocast_enabled, dtype = autocast_dtype):
            net = self.net
            feats = {1:x}
            x = net.conv1(x)
            x = net.bn1(x)
            x = net.relu(x)
            feats[2] = x 
            x = net.maxpool(x)
            x = net.layer1(x)
            feats[4] = x 
            x = net.layer2(x)
            feats[8] = x
            if self.early_exit:
                return feats
            x = net.layer3(x)
            feats[16] = x
            x = net.layer4(x)
            feats[32] = x
            return feats

    def train(self, mode=True):
        super().train(mode)
        if self.freeze_bn:
            for m in self.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eval()
                pass

class VGG19(nn.Module):
    def __init__(self, pretrained=False, amp = False, amp_dtype = torch.float16) -> None:
        super().__init__()
        self.layers = nn.ModuleList(tvm.vgg19_bn(pretrained=pretrained).features[:40])
        self.amp = amp
        self.amp_dtype = amp_dtype

    def forward(self, x, **kwargs):
        autocast_device, autocast_enabled, autocast_dtype = get_autocast_params(x.device, self.amp, self.amp_dtype)
        with torch.autocast(device_type=autocast_device, enabled=autocast_enabled, dtype = autocast_dtype):
            feats = {}
            scale = 1
            for layer in self.layers:
                if isinstance(layer, nn.MaxPool2d):
                    feats[scale] = x
                    scale = scale*2
                x = layer(x)
            return feats

class CNNandDinov2(nn.Module):
    def __init__(self, cnn_kwargs = None, amp = False, use_vgg = False, dinov2_weights = None, amp_dtype = torch.float16):
        super().__init__()
        if dinov2_weights is None:
            dinov2_weights = torch.hub.load_state_dict_from_url("https://dl.fbaipublicfiles.com/dinov2/dinov2_vitl14/dinov2_vitl14_pretrain.pth", map_location="cpu")
        from .transformer import vit_large
        vit_kwargs = dict(img_size= 518,
            patch_size= 14,
            init_values = 1.0,
            ffn_layer = "mlp",
            block_chunks = 0,
        )

        dinov2_vitl14 = vit_large(**vit_kwargs).eval()
        dinov2_vitl14.load_state_dict(dinov2_weights)
        cnn_kwargs = cnn_kwargs if cnn_kwargs is not None else {}
        if not use_vgg:
            self.cnn = ResNet50(**cnn_kwargs)
        else:
            self.cnn = VGG19(**cnn_kwargs)
        self.amp = amp
        self.amp_dtype = amp_dtype
        if self.amp:
            dinov2_vitl14 = dinov2_vitl14.to(self.amp_dtype)
        self.dinov2_vitl14 = [dinov2_vitl14] # ugly hack to not show parameters to DDP
    
    
    def train(self, mode: bool = True):
        return self.cnn.train(mode)
    
    def forward(self, x, upsample = False):
        B,C,H,W = x.shape
        feature_pyramid = self.cnn(x)
        
        if not upsample:
            with torch.no_grad():
                if self.dinov2_vitl14[0].device != x.device:
                    self.dinov2_vitl14[0] = self.dinov2_vitl14[0].to(x.device).to(self.amp_dtype)
                dinov2_features_16 = self.dinov2_vitl14[0].forward_features(x.to(self.amp_dtype))
                features_16 = dinov2_features_16['x_norm_patchtokens'].permute(0,2,1).reshape(B,1024,H//14, W//14)
                del dinov2_features_16
                feature_pyramid[16] = features_16
        return feature_pyramid


import pytest
import torch

from easy_local_features.feature.baseline_orb import ORB_baseline
from easy_local_features.feature.detector_ensemble import EnsembleDetector, EnsembleDetectorConfig
from easy_local_features.utils import io, ops
import torch.nn.functional as F


def _count_from_detector(det, img):
    k = det.detect(img)
    if k.ndim == 3 and k.shape[0] == 1:
        return k.shape[1]
    return k.shape[0]


@pytest.mark.parametrize("num_detectors", [1, 2])
def test_ensemble_detect_single_counts(num_detectors):
    img0 = io.fromPath("tests/assets/megadepth0.jpg", gray=False, batch=False)

    # Build detectors (use different top_k to diversify)
    dets = [ORB_baseline({"top_k": 256})]
    if num_detectors == 2:
        dets.append(ORB_baseline({"top_k": 128}))

    cfg = EnsembleDetectorConfig(deduplicate=False, sort=True, max_keypoints=None)
    ens = EnsembleDetector(dets, cfg)

    # Expected total count is the sum of per-detector counts
    expected_total = sum(_count_from_detector(d, img0) for d in dets)

    kps_ens = ens.detect(img0)
    assert isinstance(kps_ens, torch.Tensor)
    assert kps_ens.ndim == 3 and kps_ens.shape[0] == 1
    assert kps_ens.shape[-1] == 2
    assert kps_ens.shape[1] == expected_total


@pytest.mark.parametrize("num_detectors", [1, 2])
def test_ensemble_detect_batch_shapes(num_detectors):
    img0 = io.fromPath("tests/assets/megadepth0.jpg", gray=False, batch=False)
    img1 = io.fromPath("tests/assets/megadepth1.jpg", gray=False, batch=False)

    dets = [ORB_baseline({"top_k": 192})]
    if num_detectors == 2:
        dets.append(ORB_baseline({"top_k": 96}))

    cfg = EnsembleDetectorConfig(deduplicate=False, sort=True, max_keypoints=None)
    ens = EnsembleDetector(dets, cfg)

    # Resize to a common short edge to allow stacking
    img0_r, _ = ops.resize_short_edge(img0, 320)
    img1_r, _ = ops.resize_short_edge(img1, 320)
    # Pad right to the same width
    _, _, H0, W0 = img0_r.unsqueeze(0).shape
    _, _, H1, W1 = img1_r.unsqueeze(0).shape
    Wmax = max(W0, W1)
    if W0 < Wmax:
        pad0 = Wmax - W0
        img0_r = F.pad(img0_r, (0, pad0, 0, 0))
    if W1 < Wmax:
        pad1 = Wmax - W1
        img1_r = F.pad(img1_r, (0, pad1, 0, 0))
    batch = torch.stack([img0_r, img1_r], 0)
    kps_b = ens.detect(batch)
    assert kps_b.ndim == 3 and kps_b.shape[0] == 2
    assert kps_b.shape[-1] == 2
    # N dimension is padded to the max per-image keypoint count internally
    assert kps_b.shape[1] > 0


def test_ensemble_max_keypoints_cap():
    img0 = io.fromPath("tests/assets/megadepth0.jpg", gray=False, batch=False)

    d1 = ORB_baseline({"top_k": 512})
    d2 = ORB_baseline({"top_k": 512})
    cap = 100
    ens = EnsembleDetector([d1, d2], EnsembleDetectorConfig(deduplicate=False, sort=False, max_keypoints=cap))
    kps = ens.detect(img0)
    assert kps.ndim == 3 and kps.shape[0] == 1
    assert kps.shape[-1] == 2
    assert kps.shape[1] <= cap


def test_ensemble_deduplicate_with_identical_detectors():
    img0 = io.fromPath("tests/assets/megadepth0.jpg", gray=False, batch=False)

    # Two identical ORB detectors -> after dedup, should match single-detector count
    d1 = ORB_baseline({"top_k": 300})
    d2 = ORB_baseline({"top_k": 300})
    _ = _count_from_detector(d1, img0)

    # Without dedup
    ens_no = EnsembleDetector([d1, d2], EnsembleDetectorConfig(deduplicate=False, sort=True, max_keypoints=None))
    kps_no = ens_no.detect(img0)
    n_no = kps_no.shape[1]

    # With dedup
    ens = EnsembleDetector([d1, d2], EnsembleDetectorConfig(deduplicate=True, sort=True, max_keypoints=None))
    kps = ens.detect(img0)
    assert kps.ndim == 3 and kps.shape[0] == 1
    assert kps.shape[-1] == 2
    n_dedup = kps.shape[1]

    # Dedup should not increase count
    assert n_dedup <= n_no
    # And should produce at least one keypoint
    assert n_dedup > 0



# Copyright (C) 2022-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).


# --------------------------------------------------------
# CroCo model during pretraining
# --------------------------------------------------------


import torch
import torch.nn as nn

torch.backends.cuda.matmul.allow_tf32 = True  # for gpu >= Ampere and pytorch >= 1.12
from functools import partial  # noqa: E402

from .blocks import Block, DecoderBlock, PatchEmbed  # noqa: E402
from .masking import RandomMask  # noqa: E402
from .pos_embed import RoPE2D, get_2d_sincos_pos_embed  # noqa: E402


class CroCoNet(nn.Module):
    def __init__(
        self,
        img_size=224,  # input image size
        patch_size=16,  # patch_size
        mask_ratio=0.9,  # ratios of masked tokens
        enc_embed_dim=768,  # encoder feature dimension
        enc_depth=12,  # encoder depth
        enc_num_heads=12,  # encoder number of heads in the transformer block
        dec_embed_dim=512,  # decoder feature dimension
        dec_depth=8,  # decoder depth
        dec_num_heads=16,  # decoder number of heads in the transformer block
        mlp_ratio=4,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        norm_im2_in_dec=True,  # whether to apply normalization of the 'memory' = (second image) in the decoder
        pos_embed="cosine",  # positional embedding (either cosine or RoPE100)
    ):
        super(CroCoNet, self).__init__()

        # patch embeddings  (with initialization done as in MAE)
        self._set_patch_embed(img_size, patch_size, enc_embed_dim)

        # mask generations
        self._set_mask_generator(self.patch_embed.num_patches, mask_ratio)

        self.pos_embed = pos_embed
        if pos_embed == "cosine":
            # positional embedding of the encoder
            enc_pos_embed = get_2d_sincos_pos_embed(
                enc_embed_dim, int(self.patch_embed.num_patches**0.5), n_cls_token=0
            )
            self.register_buffer("enc_pos_embed", torch.from_numpy(enc_pos_embed).float())
            # positional embedding of the decoder
            dec_pos_embed = get_2d_sincos_pos_embed(
                dec_embed_dim, int(self.patch_embed.num_patches**0.5), n_cls_token=0
            )
            self.register_buffer("dec_pos_embed", torch.from_numpy(dec_pos_embed).float())
            # pos embedding in each block
            self.rope = None  # nothing for cosine
        elif pos_embed.startswith("RoPE"):  # eg RoPE100
            self.enc_pos_embed = None  # nothing to add in the encoder with RoPE
            self.dec_pos_embed = None  # nothing to add in the decoder with RoPE
            if RoPE2D is None:
                raise ImportError("Cannot find cuRoPE2D, please install it following the README instructions")
            freq = float(pos_embed[len("RoPE") :])
            self.rope = RoPE2D(freq=freq)
        else:
            raise NotImplementedError("Unknown pos_embed " + pos_embed)

        # transformer for the encoder
        self.enc_depth = enc_depth
        self.enc_embed_dim = enc_embed_dim
        self.enc_blocks = nn.ModuleList(
            [
                Block(enc_embed_dim, enc_num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer, rope=self.rope)
                for i in range(enc_depth)
            ]
        )
        self.enc_norm = norm_layer(enc_embed_dim)

        # masked tokens
        self._set_mask_token(dec_embed_dim)

        # decoder
        self._set_decoder(
            enc_embed_dim, dec_embed_dim, dec_num_heads, dec_depth, mlp_ratio, norm_layer, norm_im2_in_dec
        )

        # prediction head
        self._set_prediction_head(dec_embed_dim, patch_size)

        # initializer weights
        self.initialize_weights()

    def _set_patch_embed(self, img_size=224, patch_size=16, enc_embed_dim=768):
        self.patch_embed = PatchEmbed(img_size, patch_size, 3, enc_embed_dim)

    def _set_mask_generator(self, num_patches, mask_ratio):
        self.mask_generator = RandomMask(num_patches, mask_ratio)

    def _set_mask_token(self, dec_embed_dim):
        self.mask_token = nn.Parameter(torch.zeros(1, 1, dec_embed_dim))

    def _set_decoder(
        self, enc_embed_dim, dec_embed_dim, dec_num_heads, dec_depth, mlp_ratio, norm_layer, norm_im2_in_dec
    ):
        self.dec_depth = dec_depth
        self.dec_embed_dim = dec_embed_dim
        # transfer from encoder to decoder
        self.decoder_embed = nn.Linear(enc_embed_dim, dec_embed_dim, bias=True)
        # transformer for the decoder
        self.dec_blocks = nn.ModuleList(
            [
                DecoderBlock(
                    dec_embed_dim,
                    dec_num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=True,
                    norm_layer=norm_layer,
                    norm_mem=norm_im2_in_dec,
                    rope=self.rope,
                )
                for i in range(dec_depth)
            ]
        )
        # final norm layer
        self.dec_norm = norm_layer(dec_embed_dim)

    def _set_prediction_head(self, dec_embed_dim, patch_size):
        self.prediction_head = nn.Linear(dec_embed_dim, patch_size**2 * 3, bias=True)

    def initialize_weights(self):
        # patch embed
        self.patch_embed._init_weights()
        # mask tokens
        if self.mask_token is not None:
            torch.nn.init.normal_(self.mask_token, std=0.02)
        # linears and layer norms
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def _encode_image(self, image, do_mask=False, return_all_blocks=False):
        """
        image has B x 3 x img_size x img_size
        do_mask: whether to perform masking or not
        return_all_blocks: if True, return the features at the end of every block
                           instead of just the features from the last block (eg for some prediction heads)
        """
        # embed the image into patches  (x has size B x Npatches x C)
        # and get position if each return patch (pos has size B x Npatches x 2)
        x, pos = self.patch_embed(image)
        # add positional embedding without cls token
        if self.enc_pos_embed is not None:
            x = x + self.enc_pos_embed[None, ...]
        # apply masking
        B, N, C = x.size()
        if do_mask:
            masks = self.mask_generator(x)
            x = x[~masks].view(B, -1, C)
            posvis = pos[~masks].view(B, -1, 2)
        else:
            B, N, C = x.size()
            masks = torch.zeros((B, N), dtype=bool)
            posvis = pos
        # now apply the transformer encoder and normalization
        if return_all_blocks:
            out = []
            for blk in self.enc_blocks:
                x = blk(x, posvis)
                out.append(x)
            out[-1] = self.enc_norm(out[-1])
            return out, pos, masks
        else:
            for blk in self.enc_blocks:
                x = blk(x, posvis)
            x = self.enc_norm(x)
            return x, pos, masks

    def _decoder(self, feat1, pos1, masks1, feat2, pos2, return_all_blocks=False):
        """
        return_all_blocks: if True, return the features at the end of every block
                           instead of just the features from the last block (eg for some prediction heads)

        masks1 can be None => assume image1 fully visible
        """
        # encoder to decoder layer
        visf1 = self.decoder_embed(feat1)
        f2 = self.decoder_embed(feat2)
        # append masked tokens to the sequence
        B, Nenc, C = visf1.size()
        if masks1 is None:  # downstreams
            f1_ = visf1
        else:  # pretraining
            Ntotal = masks1.size(1)
            f1_ = self.mask_token.repeat(B, Ntotal, 1).to(dtype=visf1.dtype)
            f1_[~masks1] = visf1.view(B * Nenc, C)
        # add positional embedding
        if self.dec_pos_embed is not None:
            f1_ = f1_ + self.dec_pos_embed
            f2 = f2 + self.dec_pos_embed
        # apply Transformer blocks
        out = f1_
        out2 = f2
        if return_all_blocks:
            _out, out = out, []
            for blk in self.dec_blocks:
                _out, out2 = blk(_out, out2, pos1, pos2)
                out.append(_out)
            out[-1] = self.dec_norm(out[-1])
        else:
            for blk in self.dec_blocks:
                out, out2 = blk(out, out2, pos1, pos2)
            out = self.dec_norm(out)
        return out

    def patchify(self, imgs):
        """
        imgs: (B, 3, H, W)
        x: (B, L, patch_size**2 *3)
        """
        p = self.patch_embed.patch_size[0]
        assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0

        h = w = imgs.shape[2] // p
        x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
        x = torch.einsum("nchpwq->nhwpqc", x)
        x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))

        return x

    def unpatchify(self, x, channels=3):
        """
        x: (N, L, patch_size**2 *channels)
        imgs: (N, 3, H, W)
        """
        patch_size = self.patch_embed.patch_size[0]
        h = w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]
        x = x.reshape(shape=(x.shape[0], h, w, patch_size, patch_size, channels))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], channels, h * patch_size, h * patch_size))
        return imgs

    def forward(self, img1, img2):
        """
        img1: tensor of size B x 3 x img_size x img_size
        img2: tensor of size B x 3 x img_size x img_size

        out will be    B x N x (3*patch_size*patch_size)
        masks are also returned as B x N just in case
        """
        # encoder of the masked first image
        feat1, pos1, mask1 = self._encode_image(img1, do_mask=True)
        # encoder of the second image
        feat2, pos2, _ = self._encode_image(img2, do_mask=False)
        # decoder
        decfeat = self._decoder(feat1, pos1, mask1, feat2, pos2)
        # prediction head
        out = self.prediction_head(decfeat)
        # get target
        target = self.patchify(img1)
        return out, mask1, target



import torch
from kornia.feature import DeDoDe
from omegaconf import OmegaConf

from ..utils import ops
from .basemodel import BaseExtractor
from typing import TypedDict


class DeDoDeConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    detector_weights: str
    descriptor_weights: str
    amp_dtype: str


def dual_softmax_matcher(desc_A, desc_B, inv_temperature=1, normalize=False):
    if len(desc_A.shape) < 3:
        desc_A, desc_B = desc_A[None], desc_B[None]
    if normalize:
        desc_A = desc_A / desc_A.norm(dim=-1, keepdim=True)
        desc_B = desc_B / desc_B.norm(dim=-1, keepdim=True)
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    else:
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    P = corr.softmax(dim=-2) * corr.softmax(dim=-1)
    return P


class DualSoftMaxMatcher(torch.nn.Module):
    @torch.inference_mode()
    def forward(
        self, keypoints_A, descriptions_A, keypoints_B, descriptions_B, normalize=False, inv_temp=1, threshold=0.0
    ):
        if isinstance(descriptions_A, list):
            matches = [
                self.forward(
                    k_A[None],
                    d_A[None],
                    k_B[None],
                    d_B[None],
                    normalize=normalize,
                    inv_temp=inv_temp,
                    threshold=threshold,
                )
                for k_A, d_A, k_B, d_B in zip(keypoints_A, descriptions_A, keypoints_B, descriptions_B)
            ]
            matches_A = torch.cat([m[0] for m in matches])
            matches_B = torch.cat([m[1] for m in matches])
            inds = torch.cat([m[2] + b for b, m in enumerate(matches)])
            return matches_A, matches_B, inds

        P = dual_softmax_matcher(
            descriptions_A,
            descriptions_B,
            normalize=normalize,
            inv_temperature=inv_temp,
        )
        inds = torch.nonzero(
            (P == P.max(dim=-1, keepdim=True).values) * (P == P.max(dim=-2, keepdim=True).values) * (P > threshold)
        )
        batch_inds = inds[:, 0]
        matches_A = keypoints_A[batch_inds, inds[:, 1]]
        matches_B = keypoints_B[batch_inds, inds[:, 2]]
        return matches_A, matches_B, batch_inds


class DeDoDe_baseline(BaseExtractor):
    # detector_weights: The weights to load for the detector. One of 'L-upright', 'L-C4', 'L-SO2'.
    # descriptor_weights: The weights to load for the descriptor. One of 'B-upright', 'B-C4', 'B-SO2', 'G-upright', 'G-C4'.
    default_conf: DeDoDeConfig = {
        "model_name": "dedode",
        "top_k": 2048,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "detector_weights": "L-upright",  # 'L-upright', 'L-C4', 'L-SO2'
        "descriptor_weights": "B-upright",  # 'B-upright', 'B-C4', 'B-SO2', 'G-upright', 'G-C4'
        "amp_dtype": "float16",
    }

    def __init__(self, conf: DeDoDeConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.DEV = torch.device("cpu")
        self.model = DeDoDe.from_pretrained(
            detector_weights=conf.detector_weights,
            descriptor_weights=conf.descriptor_weights,
        )
        self.model.eval()
        self.matcher = DualSoftMaxMatcher()
        self.top_kps = conf.top_k

    def detectAndCompute(self, image, return_dict=False):
        image = ops.prepareImage(image, imagenet=True).to(self.DEV)
        keypoints, scores, descriptors = self.model(image, n=self.top_kps, apply_imagenet_normalization=False)

        if return_dict:
            return {"keypoints": keypoints, "descriptors": descriptors, "scores": scores}

        return keypoints, descriptors

    def detect(self, img):
        return self.detectAndCompute(img)[0]

    def compute(self, image, keypoints):
        raise NotImplementedError

    def to(self, device):
        self.model.to(device)
        self.DEV = device
        self.model.eval()

    @torch.inference_mode()
    def match(self, image1, image2):
        kp0, desc0 = self.detectAndCompute(image1)
        kp1, desc1 = self.detectAndCompute(image2)

        matches_A, matches_B, batch_ids = self.matcher(
            kp0, desc0, kp1, desc1, normalize=True, inv_temp=20, threshold=0.1
        )

        # Ensure keypoints are detached and on CPU
        return {
            "mkpts0": matches_A.detach().cpu(),
            "mkpts1": matches_B.detach().cpu(),
            "batch_ids": batch_ids.detach().cpu() if isinstance(batch_ids, torch.Tensor) else batch_ids,
        }

    @property
    def has_detector(self):
        return True



import os, tqdm
# TODO: add output path

def retrieve(dataset="hpatches"):

    datasets_links = {
            "Alamo":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Alamo.tar", ## VALIDATION
            "EllisIsland":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Ellis_Island.tar", ## TEST
            "MadridMetropolis":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Madrid_Metropolis.tar",  ## ALL BELOW TRAIN
            "MontrealNotreDame":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Montreal_Notre_Dame.tar",
            "NYC_Library":"http://landmark.cs.cornell.edu/projects/1dsfm/images.NYC_Library.tar",
            "PiazzadelPopolo":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Piazza_del_Popolo.tar",
            "Piccadilly":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Piccadilly.tar",
            "RomanForum":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Roman_Forum.tar",
            "TowerofLondon":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Tower_of_London.tar",
            "Trafalgar":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Trafalgar.tar",
            "UnionSquare":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Union_Square.tar",
            "ViennaCathedral":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Vienna_Cathedral.tar",
            "Yorkminster":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Yorkminster.tar",
            "Gendarmenmarkt":"http://landmark.cs.cornell.edu/projects/1dsfm/images.Gendarmenmarkt.tar",
            }

    os.makedirs('downloads', exist_ok=True)

    # download hpatches http://icvl.ee.ic.ac.uk/vbalnt/hpatches/hpatches-sequences-release.tar.gz
    
    if dataset == "hpatches":
        print("Downloading hpatches")
        if os.system("wget -nc http://icvl.ee.ic.ac.uk/vbalnt/hpatches/hpatches-sequences-release.tar.gz -O downloads/hpatches-sequences-release.tar.gz") or \
        os.system("tar -xv --skip-old-files -f downloads/hpatches-sequences-release.tar.gz -C downloads/"):
            raise RuntimeError('Failed to parse dataset.')

        # os.system("rm downloads/hpatches-sequences-release.tar.gz") ### If you want to save space uncomment this line to remove the tar files

    elif dataset == "sfm":
        os.makedirs("downloads/sfm", exist_ok=True)
        for key, item in tqdm.tqdm( datasets_links.items() ):
            if not os.path.isdir("downloads/sfm/" + key):
                print("Downloading " + key )
                if os.system("wget -nc "+ item + " -O downloads/sfm" + key+".tar") or \
                os.system("tar -xv --skip-old-files -f downloads/sfm" + key+".tar -C downloads/sfm" ):
                    raise RuntimeError('Failed to parse dataset.')

                os.system("rm downloads/sfm/" + key+".tar") ### If you want to save space uncomment this line to remove the tar files

    else:
        raise RuntimeError('Dataset not supported.')
    
def sfm():
    retrieve("sfm")

def hpatches():
    retrieve("hpatches")



import torch, abc, sys

# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

# the class/object below is there for making type annotations like
# def my_function(args) -> NpArray[OutputType]
if sys.version_info >= (3, 7):
    class NpArray:
        def __class_getitem__(self, arg):
            pass
else:
    # 3.6 and below don't support __class_getitem__
    class _NpArray:
        def __getitem__(self, _idx):
            pass
    
    NpArray = _NpArray()

class Features:
    # @dimchecked
    def __init__(self, kp, desc, kp_logp):
        assert kp.device == desc.device
        assert kp.device == kp_logp.device

        self.kp      = kp
        self.desc    = desc
        self.kp_logp = kp_logp

    @property
    def n(self):
        return self.kp.shape[0]

    @property
    def device(self):
        return self.kp.device

    def detached_and_grad_(self):
        return Features(
            self.kp,
            self.desc.detach().requires_grad_(),
            self.kp_logp.detach().requires_grad_(),
        )

    def requires_grad_(self, is_on):
        self.desc.requires_grad_(is_on)
        self.kp_logp.requires_grad_(is_on)

    def grad_tensors(self):
        return [self.desc, self.kp_logp]

    def to(self, *args, **kwargs):
        return Features(
            self.kp.to(*args, **kwargs),
            self.desc.to(*args, **kwargs),
            self.kp_logp.to(*args, **kwargs) if self.kp_logp is not None else None,
        )

class MatchDistribution(abc.ABC):
    @abc.abstractmethod
    def sample(self):
        pass

    @abc.abstractmethod
    def mle(self):
        pass

    @abc.abstractmethod
    def dense_logp(self):
        pass

    @abc.abstractmethod
    def dense_p(self):
        pass

    @abc.abstractmethod
    def features_1(self) -> Features:
        pass

    @abc.abstractmethod
    def features_2(self) -> Features:
        pass

    @property
    def shape(self):
        return self.features_1().kp.shape[0], self.features_2().kp.shape[1]

    def matched_pairs(self, mle=False):
        matches = self.mle() if mle else self.sample()

        return MatchedPairs(
            self.features_1().kp,
            self.features_2().kp,
            matches,    
        )


class MatchedPairs:
    # @dimchecked
    def __init__(self, kps1, kps2, matches):
        self.kps1    = kps1
        self.kps2    = kps2
        self.matches = matches

    def to(self, *args, **kwargs):
        return MatchedPairs(
            self.kps1.to(*args, **kwargs),
            self.kps2.to(*args, **kwargs),
            self.matches.to(*args, **kwargs),
        )



'''
    This script implements color + geometric transformations using Kornia
    Given a dataset of random images, we apply color transformations, 
    homography warps and also TPS warps.

    @author: Guilherme Potje
    @cite: @INPROCEEDINGS{potje2023cvpr,
            author={Guilherme {Potje} and and Felipe {Cadar} and Andre {Araujo} and Renato {Martins} and Erickson R. {Nascimento}},
            booktitle={2023 IEEE / CVF Computer Vision and Pattern Recognition (CVPR)}, 
            title={Enhancing Deformable Local Features by Jointly Learning to Detect and Describe Keypoints}, 
            year={2023}
    }
'''

import torch
from torch import nn
from torch.utils.data import Dataset
import torch.utils.data as data
from torchvision import transforms
import torch.nn.functional as F

import cv2
import kornia
import kornia.augmentation as K
from kornia.geometry.transform import get_tps_transform as findTPS
from kornia.geometry.transform import warp_points_tps, warp_image_tps

import glob
import random
import tqdm

import numpy as np
import pdb
import time

random.seed(0)
torch.manual_seed(0)

def generateRandomTPS(shape, grid = (8, 6), GLOBAL_MULTIPLIER = 0.3, prob = 0.5):

    h, w = shape
    sh, sw = h/grid[0], w/grid[1]
    src = torch.dstack(torch.meshgrid(torch.arange(0, h + sh , sh),
                         torch.arange(0, w + sw , sw), indexing="ij"))

    offsets = torch.rand(grid[0]+1, grid[1]+1, 2) - 0.5
    offsets *= torch.tensor([ sh/2, sw/2 ]).view(1, 1, 2)  * min(0.97, 3. * GLOBAL_MULTIPLIER)
    dst = src + offsets if np.random.uniform() < prob else src
    
    src, dst = src.view(1, -1, 2), dst.view(1, -1, 2)
    src = (src / torch.tensor([h,w]).view(1,1,2) ) * 2 - 1.
    dst = (dst / torch.tensor([h,w]).view(1,1,2) ) * 2 - 1.
    weights, A = findTPS(dst, src)

    return src, weights, A



def generateRandomHomography(shape, GLOBAL_MULTIPLIER = 0.3):

    theta = np.radians(np.random.normal(0, 12.0*GLOBAL_MULTIPLIER)) 

    scale = np.random.normal(0, 1.1*GLOBAL_MULTIPLIER)#0.15*GLOBAL_MULTIPLIER)
    if scale < 0.0: # get the right part of the gaussian
        scale = 1.0/(1. + abs(scale))
    else:
        scale = 1. + scale

    tx , ty = -shape[1]/2.0 , -shape[0]/2.0 
    txn, tyn = np.random.normal(0, 80.0*GLOBAL_MULTIPLIER, 2) #translation error
    c, s = np.cos(theta), np.sin(theta)

    sx , sy = np.random.normal(0,0.6*GLOBAL_MULTIPLIER,2)
    p1 , p2 = np.random.normal(0,0.006*GLOBAL_MULTIPLIER,2)
    if np.random.uniform() > 0.5:
        scale_r = np.random.uniform(max(0.2, 0.8 - GLOBAL_MULTIPLIER), 1.)
    else:
        scale_r = np.random.uniform(1., min(4., 1.2 + 3.*GLOBAL_MULTIPLIER))

    H_t = np.array(((1,0, tx), (0, 1, ty), (0,0,1))) #t

    H_r = np.array(((c,-s, 0), (s, c, 0), (0,0,1))) #rotation,
    H_a = np.array(((1,sy, 0), (sx, 1, 0), (0,0,1))) # affine
    H_p = np.array(((1, 0, 0), (0 , 1, 0), (p1,p2,1))) # projective

    H_s = np.array(((scale,0, 0), (0, scale * scale_r, 0), (0,0,1))) #scale
    H_b = np.array(((1.0,0,-tx +txn), (0, 1, -ty + tyn), (0,0,1))) #t_back,

    #H = H_e * H_s * H_a * H_p
    H = np.dot(np.dot(np.dot(np.dot(np.dot(H_b,H_s),H_p),H_a),H_r),H_t)

    return H

class AugmentationPipe(nn.Module):
    def __init__(
                self, device,
                img_dir=None,
                load_dataset = False,
                warp_resolution = (448, 448),
                out_resolution = (400, 300),
                max_num_imgs = 40,
                num_test_imgs = 200,
                batch_size = 6,
                sample_img = None
                ):
        super(AugmentationPipe, self).__init__()
        self.half = 16
        self.device = device
        self.dims = warp_resolution
        self.batch_size = batch_size
        self.out_resolution = out_resolution 
        self.dims_t = torch.tensor([int(self.dims[0]*0.8) - int(self.dims[0]*0.2) -1,
                                    int(self.dims[1]*0.8) - int(self.dims[1]*0.2) -1]).float().to(device).view(1,1,2)
        self.dims_s = torch.tensor([ self.dims_t[0,0,0] / out_resolution[0],
                                     self.dims_t[0,0,1] / out_resolution[1]]).float().to(device).view(1,1,2) 

        self.aug_list = kornia.augmentation.ImageSequential(
                        kornia.augmentation.RandomChannelShuffle(p=0.5),
                        kornia.augmentation.ColorJitter(0.2, 0.2, 0.2, 0.2, p=1.),
                        kornia.augmentation.RandomEqualize(p = 0.5),
                        kornia.augmentation.RandomGaussianBlur(p = 0.3, sigma = (2.5, 2.5), kernel_size = (7,7))
                        )
        
        if sample_img:
            self.sample_img = cv2.imread(sample_img)
            self.sample_img = cv2.resize(self.sample_img, self.dims)

        if load_dataset and img_dir is None:
            raise RuntimeError('Error: no dataset path given!')

        if load_dataset:
            self.all_imgs = glob.glob(img_dir)
            random.shuffle(self.all_imgs)

            print('Found a total of ', len(self.all_imgs), ' images for training..')

            if len(self.all_imgs) - num_test_imgs < max_num_imgs:
                raise RuntimeError('Error: test set overlaps with training set! Decrease number of test imgs')

            train = []
            fast = cv2.FastFeatureDetector_create(30)
            for p in tqdm.tqdm(self.all_imgs[:max_num_imgs], desc='loading train'):
                im = cv2.imread(p)
                halfH, halfW = im.shape[0]//2, im.shape[1]//2
                if halfH > halfW:
                    im = np.rot90(im)
                    halfH, halfW = halfW, halfH
                im = im[halfH-self.dims[1]//2:halfH+self.dims[1]//2, halfW-self.dims[0]//2:halfW+self.dims[0]//2, :]
                #print (im.shape)
                if im.shape[0] != self.dims[1] or im.shape[1] != self.dims[0]:
                    #print('resizing..')
                    im = cv2.resize(im, self.dims)
                if len(fast.detect(im)) > 1_000:
                    train.append(np.copy(im))

            self.train = train
            
            self.test = [
                        cv2.resize(cv2.imread(p), self.dims)                 
                        for p in tqdm.tqdm(self.all_imgs[-num_test_imgs:],
                                           desc='loading test')
                        ] 

            self.TPS = True

    def load_image(self, path):
        im = cv2.imread(path)
        halfH, halfW = im.shape[0]//2, im.shape[1]//2
        if halfH > halfW:
            im = np.rot90(im)
            halfH, halfW = halfW, halfH
        im = im[halfH-self.dims[1]//2:halfH+self.dims[1]//2, halfW-self.dims[0]//2:halfW+self.dims[0]//2, :]
        if im.shape[0] != self.dims[1] or im.shape[1] != self.dims[0]:
            im = cv2.resize(im, self.dims)

        return im

    def norm_pts_grid(self, x):
        if len(x.size()) == 2:
            return (x.view(1,-1,2) * self.dims_s / self.dims_t) * 2. - 1 
        return (x * self.dims_s / self.dims_t) * 2. - 1

    def denorm_pts_grid(self, x):
        if len(x.size()) == 2:
            return ((x.view(1,-1,2) + 1) / 2.) / self.dims_s * self.dims_t
        return ((x+1) / 2.) / self.dims_s * self.dims_t

    def rnd_kps(self, shape, n = 256):
        h, w = shape
        kps = torch.rand(size = (3,n)).to(self.device)
        kps[0,:]*=w
        kps[1,:]*=h
        kps[2,:] = 1.0

        return kps

    def warp_points(self, H, pts):
      scale = self.dims_s.view(-1,2)
      offset = torch.tensor([int(self.dims[0]*0.2), int(self.dims[1]*0.2)], device = pts.device).float()
      pts = pts*scale + offset
      pts = torch.vstack( [pts.t(), torch.ones(1, pts.shape[0], device = pts.device)])
      warped = torch.matmul(H, pts)
      warped = warped / warped[2,...] 
      warped = warped.t()[:, :2]
      return (warped - offset) / scale

    def forward(self, x, difficulty = 0.3, TPS = False, prob_deformation = 0.5, test = False):
        with torch.no_grad():
            x = (x/255.).to(self.device)
            shape = x.shape[-2:]
            h, w = shape

            #t0 = time.time()
            output = self.aug_list(x)

            #Correlated Gaussian Noise
            if np.random.uniform() > 0.5:
                noise = F.interpolate(torch.randn_like(output)*(16/255), (h//2, w//2))
                noise = F.interpolate(noise, (h, w), mode = 'bicubic')
                output = torch.clip( output + noise, 0., 1.)

            #print('done in ', time.time() - t0)
            H = torch.tensor([generateRandomHomography(shape, difficulty) for b in range(self.batch_size)],
                               dtype = torch.float32).to(self.device)
            
            output = kornia.geometry.transform.warp_perspective(output, H,
                            dsize = shape, padding_mode = 'zeros')


            #crop 20% of image boundaries each side to reduce invalid pixels after warps
            low_h = int(h * 0.2); low_w = int(w*0.2)
            high_h = int(h*0.8); high_w= int(w * 0.8)
            output = output[..., low_h:high_h, low_w:high_w]

            #apply TPS if desired:
            
            if TPS:
                src, weights, A = None, None, None
                for b in range(self.batch_size):
                    b_src, b_weights, b_A = generateRandomTPS(shape, (8,6), difficulty, prob = prob_deformation)
                    b_src, b_weights, b_A = b_src.to(self.device), b_weights.to(self.device), b_A.to(self.device)

                    if src is None:
                        src, weights, A = b_src, b_weights, b_A
                    else:
                        src = torch.cat((b_src, src))
                        weights = torch.cat((b_weights, weights))
                        A = torch.cat((b_A, A))

                #print(output.shape, src.shape, weights.shape, A.shape)
                output = warp_image_tps(output, src, weights, A)

            output = F.interpolate(output, self.out_resolution[::-1], mode = 'bilinear', align_corners = False)

        
        if TPS:
            return output, (H, src, weights, A)
        else:
            return output, H

    def get_correspondences(self, kps_target, T):
        H, H2, src, W, A = T
        undeformed  = self.denorm_pts_grid(   
                                        warp_points_tps(self.norm_pts_grid(kps_target),
                                        src, W, A) ).view(-1,2)

        warped_to_src = self.warp_points(H@torch.inverse(H2), undeformed)

        return warped_to_src
    
if __name__ == "__main__":
    
    device = torch.device('cpu')
    pipe = AugmentationPipe(device, batch_size = 1)


from .models import *
from e2cnn.nn import *

import easydict, torch
import torch.nn.functional as F

class E2CNN_wrapper(torch.nn.Module):
    def __init__(self, args):
        super(E2CNN_wrapper, self).__init__()

        config = {'model': args.model, 'flip': False, 'sgsize': None, 'fixparams': False, 'N': args.num_group,
                'F': 0.8, 'sigma': 0.6, 'J': 0, 'restrict': -1, 'deltaorth': False, 'antialias': 0.0}
        config = easydict.EasyDict(config)

        model = build_model(config, 3, 10) ## input channel, output dummy  num of classes.
        model.eval()

        self.in_type = model.in_repr if 'E2SFCNN' in args.model else model.in_type
        self.model = model
        self.G = args.num_group

    def get_in_type(self):
        return self.in_type
    
    def forward(self, x):
        return self.multiscale_forward(x)

    def multiscale_forward(self, x):
        a, b, c = self.model.features(x.tensor)  

        B, _, H, W = a.shape ## first feature shape
        _, CG, _, _ = c.shape ## last feature shape
        output = F.interpolate(c.tensor, size=(H, W), mode='bilinear', align_corners=True).reshape(B, -1, self.G, H, W)

        for ff in [b, a]:
            ff = F.interpolate(ff.tensor, size=(H, W), mode='bilinear', align_corners=True) ## spatial size align.
            ff = ff.reshape(B, -1, self.G, H, W)
            output = torch.cat([output, ff], dim=1)

        output = output.reshape(B, -1, H, W)

        return output
    
    def singlescale_forward(self, x):
        raise NotImplementedError

def build_model(config, n_inputs, n_outputs):
    # SFCNN VARIANTS
    if config.model == 'E2SFCNN':
        model = E2SFCNN(n_inputs, n_outputs, restrict=config.restrict, N=config.N, fco=config.F, J=config.J,
                        sigma=config.sigma, fix_param=config.fixparams, sgsize=config.sgsize, flip=config.flip)
    elif config.model == 'E2SFCNN_QUOT':
        model = E2SFCNN_QUOT(n_inputs, n_outputs, restrict=config.restrict, N=config.N, fco=config.F, J=config.J,
                             sigma=config.sigma, sgsize=config.sgsize, flip=config.flip)
    elif config.model == 'EXP':
        model = ExpE2SFCNN(n_inputs, n_outputs, layer_type=config.type, restrict=config.restrict, N=config.N,
                           fix_param=config.fixparams, fco=config.F, J=config.J, sigma=config.sigma,
                           deltaorth=config.deltaorth, antialias=config.antialias, sgsize=config.sgsize,
                           flip=config.flip)
    elif config.model == 'CNN':
        model = ExpCNN(n_inputs, n_outputs, fix_param=config.fixparams, deltaorth=config.deltaorth)
    elif config.model == "wrn16_8_stl":
        model = wrn16_8_stl(num_classes=n_outputs, deltaorth=config.deltaorth)
    elif config.model == "e2wrn16_8_stl":
        model = e2wrn16_8_stl(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                               deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn28_10":
        model = e2wrn28_10(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                            deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn28_7":
        model = e2wrn28_7(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                           deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn28_10R":
        model = e2wrn28_10R(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                             deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn28_7R":
        model = e2wrn28_7R(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                            deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn10_8R":
        model = e2wrn10_8R(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                            deltaorth=config.deltaorth, fixparams=config.fixparams)
    elif config.model == "e2wrn16_8R_stl":
        model = e2wrn16_8R_stl(N=config.N, r=config.restrict, num_classes=n_outputs, sigma=config.sigma, F=config.F,
                            deltaorth=config.deltaorth, fixparams=config.fixparams)                          
    else:
        raise ValueError("Model selected ({}) not recognized!".format(config.model))
    
    return model



import torch
import numpy as np
import torch.nn.functional as F

from torch.distributions import Categorical, Bernoulli

# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

from ..common.structs import NpArray, Features
from ..model.nms import nms

# @dimchecked
def select_on_last(values, indices):
    '''
    WARNING: this may be reinventing the wheel, but I don't know how to do
    it otherwise with PyTorch.

    This function uses an array of linear indices `indices` between [0, T] to
    index into `values` which has equal shape as `indices` and then one extra
    dimension of size T.
    '''
    return torch.gather(
        values,
        -1,
        indices[..., None]
    ).squeeze(-1)

# @dimchecked
def point_distribution(
    logits
):
    '''
    Implements the categorical proposal -> Bernoulli acceptance sampling
    scheme. Given a tensor of logits, performs samples on the last dimension,
    returning
        a) the proposals
        b) a binary mask indicating which ones were accepted
        c) the logp-probability of (proposal and acceptance decision)
    '''

    proposal_dist = Categorical(logits=logits)
    proposals     = proposal_dist.sample()
    proposal_logp = proposal_dist.log_prob(proposals)

    accept_logits = select_on_last(logits, proposals).squeeze(-1)

    accept_dist    = Bernoulli(logits=accept_logits)
    accept_samples = accept_dist.sample()
    accept_logp    = accept_dist.log_prob(accept_samples)
    accept_mask    = accept_samples == 1.

    logp = proposal_logp + accept_logp

    return proposals, accept_mask, logp

class Keypoints:
    '''
    A simple, temporary struct used to store keypoint detections and their
    log-probabilities. After construction, merge_with_descriptors is used to
    select corresponding descriptors from unet output.
    '''

    # @dimchecked
    def __init__(self, xys, logp):
        self.xys  = xys
        self.logp = logp

    # @dimchecked
    def merge_with_descriptors(self, descriptors) -> Features:
        '''
        Select descriptors from a dense `descriptors` tensor, at locations
        given by `self.xys`
        '''
        x, y = self.xys.T

        desc = descriptors[:, y, x].T
        desc = F.normalize(desc, dim=-1)

        return Features(self.xys.to(torch.float32), desc, self.logp)

class Detector:
    def __init__(self, window=8):
        self.window = window

    # @dimchecked
    def _tile(self, heatmap):
        '''
        Divides the heatmap `heatmap` into tiles of size (v, v) where
        v==self.window. The tiles are flattened, resulting in the last
        dimension of the output T == v * v.
        '''
        v = self.window
        b, c, h, w = heatmap.shape

        assert heatmap.shape[2] % v == 0
        assert heatmap.shape[3] % v == 0

        return heatmap.unfold(2, v, v) \
                      .unfold(3, v, v) \
                      .reshape(b, c, h // v, w // v, v*v)

    # @dimchecked
    def sample(self, heatmap) -> NpArray[Keypoints]:
        '''
            Implements the training-time grid-based sampling protocol
        '''
        v = self.window
        dev = heatmap.device
        B, _, H, W = heatmap.shape

        assert H % v == 0
        assert W % v == 0

        # tile the heatmap into [window x window] tiles and pass it to
        # the categorical distribution.
        heatmap_tiled = self._tile(heatmap).squeeze(1)
        proposals, accept_mask, logp = point_distribution(heatmap_tiled)

        # create a grid of xy coordinates and tile it as well
        cgrid = torch.stack(torch.meshgrid(
            torch.arange(H, device=dev),
            torch.arange(W, device=dev),
            indexing="ij"
        )[::-1], dim=0).unsqueeze(0)
        cgrid_tiled = self._tile(cgrid)

        # extract xy coordinates from cgrid according to indices sampled
        # before
        xys = select_on_last(
            self._tile(cgrid).repeat(B, 1, 1, 1, 1),
            # unsqueeze and repeat on the (xy) dimension to grab
            # both components from the grid
            proposals.unsqueeze(1).repeat(1, 2, 1, 1)
        ).permute(0, 2, 3, 1) # -> bhw2
         
        keypoints = []
        for i in range(B):
            mask = accept_mask[i]
            keypoints.append(Keypoints(
                xys[i][mask],
                logp[i][mask],
            ))

        return np.array(keypoints, dtype=object)

    # @dimchecked
    def nms(
        self,
        heatmap,
        n=None,
        **kwargs
    ) -> NpArray[Keypoints]:
        '''
            Inference-time nms-based detection protocol
        '''
        heatmap = heatmap.squeeze(1)
        nmsed = nms(heatmap, **kwargs)

        keypoints = []
        for b in range(heatmap.shape[0]):
            yx   = nmsed[b].nonzero(as_tuple=False)
            logp = heatmap[b][nmsed[b]]
            xy   = torch.flip(yx, (1, ))

            if n is not None:
                n_ = min(n+1, logp.numel())
                # torch.kthvalue picks in ascending order and we want to pick in
                # descending order, so we pick n-th smallest among -logp to get
                # -threshold
                minus_threshold, _indices = torch.kthvalue(-logp, n_)
                mask = logp > -minus_threshold

                xy   = xy[mask]
                logp = logp[mask]

            keypoints.append(Keypoints(xy, logp))

        return np.array(keypoints, dtype=object)



from typing import Union
import torch
from .roma_models import roma_model, tiny_roma_v1_model

weight_urls = {
    "romatch": {
        "outdoor": "https://github.com/Parskatt/storage/releases/download/roma/roma_outdoor.pth",
        "indoor": "https://github.com/Parskatt/storage/releases/download/roma/roma_indoor.pth",
    },
    "tiny_roma_v1": {
        "outdoor": "https://github.com/Parskatt/storage/releases/download/roma/tiny_roma_v1_outdoor.pth",
    },
    "dinov2": "https://dl.fbaipublicfiles.com/dinov2/dinov2_vitl14/dinov2_vitl14_pretrain.pth", #hopefully this doesnt change :D
}

def tiny_roma_v1_outdoor(device, weights = None, xfeat = None):
    if weights is None:
        weights = torch.hub.load_state_dict_from_url(
            weight_urls["tiny_roma_v1"]["outdoor"],
            map_location=device)
    if xfeat is None:
        xfeat = torch.hub.load(
            'verlab/accelerated_features', 
            'XFeat', 
            pretrained = True, 
            top_k = 4096).net

    return tiny_roma_v1_model(weights = weights, xfeat = xfeat).to(device) 

def roma_outdoor(device, weights=None, dinov2_weights=None, coarse_res: Union[int,tuple[int,int]] = 560, upsample_res: Union[int,tuple[int,int]] = 864, amp_dtype: torch.dtype = torch.float16):
    if isinstance(coarse_res, int):
        coarse_res = (coarse_res, coarse_res)
    if isinstance(upsample_res, int):    
        upsample_res = (upsample_res, upsample_res)

    if str(device) == 'cpu':
        amp_dtype = torch.float32

    assert coarse_res[0] % 14 == 0, "Needs to be multiple of 14 for backbone"
    assert coarse_res[1] % 14 == 0, "Needs to be multiple of 14 for backbone"
    
    if weights is None:
        weights = torch.hub.load_state_dict_from_url(weight_urls["romatch"]["outdoor"],
                                                     map_location=device)
    if dinov2_weights is None:
        dinov2_weights = torch.hub.load_state_dict_from_url(weight_urls["dinov2"],
                                                     map_location=device)
    model = roma_model(resolution=coarse_res, upsample_preds=True,
               weights=weights,dinov2_weights = dinov2_weights,device=device, amp_dtype=amp_dtype)
    model.upsample_res = upsample_res
    print(f"Using coarse resolution {coarse_res}, and upsample res {model.upsample_res}")
    return model

def roma_indoor(device, weights=None, dinov2_weights=None, coarse_res: Union[int,tuple[int,int]] = 560, upsample_res: Union[int,tuple[int,int]] = 864, amp_dtype: torch.dtype = torch.float16):
    if isinstance(coarse_res, int):
        coarse_res = (coarse_res, coarse_res)
    if isinstance(upsample_res, int):    
        upsample_res = (upsample_res, upsample_res)

    assert coarse_res[0] % 14 == 0, "Needs to be multiple of 14 for backbone"
    assert coarse_res[1] % 14 == 0, "Needs to be multiple of 14 for backbone"
    
    if weights is None:
        weights = torch.hub.load_state_dict_from_url(weight_urls["romatch"]["indoor"],
                                                     map_location=device)
    if dinov2_weights is None:
        dinov2_weights = torch.hub.load_state_dict_from_url(weight_urls["dinov2"],
                                                     map_location=device)
    model = roma_model(resolution=coarse_res, upsample_preds=True,
               weights=weights,dinov2_weights = dinov2_weights,device=device, amp_dtype=amp_dtype)
    model.upsample_res = upsample_res
    print(f"Using coarse resolution {coarse_res}, and upsample res {model.upsample_res}")
    return model



import torch
from einops import rearrange
from kornia.feature.sosnet import SOSNet
from omegaconf import OmegaConf

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from .basemodel import BaseExtractor, MethodType
from typing import TypedDict


class SOSNetConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    desc_dim: int


class SOSNet_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DESCRIPTOR_ONLY
    default_conf: SOSNetConfig = {
        "model_name": "sosnet",
        "top_k": 2048,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "desc_dim": 128,
    }

    def __init__(self, conf: SOSNetConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.DEV = torch.device("cpu")
        self.model = SOSNet().to(self.DEV)
        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, img, return_dict=False):
        raise NotImplementedError

    def detect(self, img, op=None):
        raise NotImplementedError

    def compute(self, img, keypoints):
        img = ops.prepareImage(img, gray=True)
        do_squeeze = False
        if len(keypoints.shape) == 2:
            keypoints = keypoints.unsqueeze(0)
            do_squeeze = True

        # patches ["B", "1", "32", "32"]
        patches = ops.crop_patches(img, keypoints, patch_size=32)

        B, N, one, PS1, PS2 = patches.shape
        patches = rearrange(patches, "B N one PS1 PS2 -> (B N) one PS1 PS2", B=B)

        desc = self.model(patches)

        desc = rearrange(desc, "(B N) D -> B N D", B=B, N=N, D=128)

        return keypoints, desc

    def to(self, device):
        self.model.to(device)
        self.DEV = device

    @property
    def has_detector(self):
        return False



from kornia.feature import LoFTR

import kornia
import torch
import functools
import numpy as np


@functools.lru_cache(maxsize=1)
def getLoFTR(pretrained="outdoor"):
    return LoFTR(pretrained=pretrained)

class LoFTR_baseline:
    def __init__(self, pretrained="outdoor"):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.matcher = getLoFTR(pretrained=pretrained).to(self.device)

    def match(self, img1, img2):

        if isinstance(img1, np.ndarray):
            img1 = kornia.utils.image_to_tensor(img1).to(self.device)
        if isinstance(img2, np.ndarray):
            img2 = kornia.utils.image_to_tensor(img2).to(self.device)

        # check if its float
        if img1.dtype != torch.float32:
            img1 = img1.float()
        if img2.dtype != torch.float32:
            img2 = img2.float()

        # check if its in range 0-1
        if img1.max() > 1.0:
            img1 = img1 / 255.0
        if img2.max() > 1.0:
            img2 = img2 / 255.0

        # batch dimension
        if len(img1.shape) == 3:
            img1 = img1.unsqueeze(0)

        if len(img2.shape) == 3:
            img2 = img2.unsqueeze(0)

        # grayscale
        if img1.shape[1] == 3:
            img1 = kornia.color.bgr_to_grayscale(img1)
        if img2.shape[1] == 3:
            img2 = kornia.color.bgr_to_grayscale(img2)

        input = {"image0": img1, "image1": img2}
        correspondences_dict = self.matcher(input)

        mkpts0 = correspondences_dict["keypoints0"].detach().cpu()
        mkpts1 = correspondences_dict["keypoints1"].detach().cpu()
        scores = correspondences_dict.get("confidence", None)
        out = {
            "mkpts0": mkpts0,
            "mkpts1": mkpts1,
        }
        if scores is not None:
            out["scores"] = scores.detach().cpu()
        return out



import os
import glob
import json
import pandas as pd
import sys
import argparse
from pathlib import Path

FOLDER_RESULTS = "./output/pose_eval/scannet"
def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("--output",  "-o", help="Output folder", default=FOLDER_RESULTS)
    parser.add_argument("--accuracy",  "-acc", action='store_true', help="Print accuracy")
    parser.add_argument("--filter",  "-f", nargs='+', help="Filter by name")
    return parser.parse_args()

if __name__ == "__main__":
    args = parse()
    
    dataset_name = 'scannet'
    all_summary_files = list(Path(args.output).glob("*/*_summary.json"))
    all_summary_files = [str(f) for f in all_summary_files]

    if len(all_summary_files) == 0:
        print(f"No summary files found at {args.output}")
        exit(1)

    filtering = []
    if args.filter:
        filtering = args.filter
        # all_summary_files = [f for f in all_summary_files if filtering in f]
        # include if in any of the filtering
        all_summary_files = [f for f in all_summary_files if any([fil in f for fil in filtering])]

    dfs = []
    names = []
    estimators = []
    metric_key = 'aucs_by_thresh'
    if args.accuracy:
        metric_key = 'accuracies_by_thresh'    
    for summary in all_summary_files:
        summary_data = json.load(open(summary, 'r'))
        if metric_key not in summary_data:
            continue
        aucs_by_thresh = summary_data[metric_key]

        estimator = 'poselib'
        if 'opencv' in summary:
            estimator = 'opencv'

        #make sure everything is float
        for thresh in aucs_by_thresh:
            for k in aucs_by_thresh[thresh]:
                if isinstance(aucs_by_thresh[thresh][k], str):
                    aucs_by_thresh[thresh][k] = float(aucs_by_thresh[thresh][k].replace(' ', ''))

        # find best threshold based on the 5, 10, 20 mAP and everything is float
        df = pd.DataFrame(aucs_by_thresh).T.astype(float)
        df['mean'] = df.mean(axis=1)
        # create a string column called estimator
        cols = df.columns.tolist()
        dfs.append(df)
        names.append(summary_data['name'])
        estimators.append(estimator)

        # print(summary_data['name'])
        # print(df)
        # print()

    # use each col as the main col to determine the best threshold
    # for col in cols:
    col = 'mean'

    final_df = pd.DataFrame()
    # add cols
    final_df['name'] = names
    final_df['best_thresh'] = ''
    final_df['estimator'] = estimators
    final_df[cols] = -1.0

    for df, name, estimator in zip(dfs, names, estimators):
        best_thresh = df[col].idxmax()
        best_results = df.loc[best_thresh]

        # final_df.loc[final_df['name'] == name, 'best_thresh'] = best_thresh
        # final_df.loc[final_df['name'] == name and final_df['estimator'] == estimator, 'best_thresh'] = best_thresh
        # for _col in cols:
            # final_df.loc[final_df['name'] == name, _col] = best_results[_col]
        
        # now update the best_thresh based on the estimator
        final_df.loc[(final_df['name'] == name) & (final_df['estimator'] == estimator), 'best_thresh'] = best_thresh
        for _col in cols:
            final_df.loc[(final_df['name'] == name) & (final_df['estimator'] == estimator), _col] = best_results[_col]

    # sort by mean
    final_df = final_df.sort_values(by=['mean'])
    # reset index
    final_df = final_df.reset_index(drop=True)

    # drop estimator column
    final_df = final_df.drop(columns=['estimator'])

    # set max float precision to 1
    final_df = final_df.round(1)

    print(f"Dataset: {dataset_name}")
    print(f"Sorting by {col}")
    print(final_df)
    print()

    final_df.to_csv(os.path.join(FOLDER_RESULTS, f"{dataset_name}_{col}.csv"), index=False)




'''=================================================
@Project -> File   r2d2 -> extract_localization
@IDE    PyCharm
@Author fx221@cam.ac.uk
@Date   13/07/2022 09:59
=================================================='''

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as tvf

RGB_mean = [0.485, 0.456, 0.406]
RGB_std = [0.229, 0.224, 0.225]

norm_RGB = tvf.Compose([tvf.Normalize(mean=RGB_mean, std=RGB_std)])

def simple_nms(scores, nms_radius: int):
    """ Fast Non-maximum suppression to remove nearby points """
    assert (nms_radius >= 0)

    def max_pool(x):
        return torch.nn.functional.max_pool2d(
            x, kernel_size=nms_radius * 2 + 1, stride=1, padding=nms_radius)

    zeros = torch.zeros_like(scores)
    max_mask = scores == max_pool(scores)
    for _ in range(2):
        supp_mask = max_pool(max_mask.float()) > 0
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == max_pool(supp_scores)
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)

def remove_borders(keypoints, scores, border: int, height: int, width: int):
    """ Removes keypoints too close to the border """
    mask_h = (keypoints[:, 0] >= border) & (keypoints[:, 0] < (height - border))
    mask_w = (keypoints[:, 1] >= border) & (keypoints[:, 1] < (width - border))
    mask = mask_h & mask_w
    return keypoints[mask], scores[mask]


def top_k_keypoints(keypoints, scores, k: int):
    if k >= len(keypoints):
        return keypoints, scores
    scores, indices = torch.topk(scores, k, dim=0)
    return keypoints[indices], scores


def sample_descriptors(keypoints, descriptors, s: int = 8):
    """ Interpolate descriptors at keypoint locations """
    b, c, h, w = descriptors.shape
    keypoints = keypoints - s / 2 + 0.5
    keypoints /= torch.tensor([(w * s - s / 2 - 0.5), (h * s - s / 2 - 0.5)],
                              ).to(keypoints)[None]
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
    args = {'align_corners': True} if torch.__version__ >= '1.3' else {}
    descriptors = torch.nn.functional.grid_sample(
        descriptors, keypoints.view(b, 1, -1, 2), mode='bilinear', **args)
    descriptors = torch.nn.functional.normalize(
        descriptors.reshape(b, c, -1), p=2, dim=1)
    return descriptors


def extractor(scores, descs, nms_radius=4, conf_th=0.001, remove_borders=8, max_keypoints=1000, scale_factor=4):
    scores = simple_nms(scores, nms_radius)

    # Extract keypoints
    keypoints = [
        torch.nonzero(s > conf_th)
        for s in scores]
    scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]

    # Discard keypoints near the image borders
    keypoints, scores = list(zip(*[
        remove_borders(k, s, remove_borders, scores.shape[2], scores.shape[3])
        for k, s in zip(keypoints, scores)]))

    # Keep the k keypoints with highest score
    if max_keypoints >= 0:
        keypoints, scores = list(zip(*[
            top_k_keypoints(k, s, max_keypoints)
            for k, s in zip(keypoints, scores)]))

    # Convert (h, w) to (x, y)
    keypoints = [torch.flip(k, [1]).float() for k in keypoints]
    # Extract descriptors
    descriptors = [sample_descriptors(k[None], d[None], scale_factor)[0]
                   for k, d in zip(keypoints, descs)]

    return keypoints, descriptors


def extract_resnet_return(model, img, conf_th=0.001,
                          mask=None,
                          topK=-1,
                          **kwargs):

    img = norm_RGB(img.squeeze())
    img = img[None]

    B, one, H, W = img.shape

    all_pts = []
    all_descs = []

    if 'scales' in kwargs.keys():
        scales = kwargs.get('scales')
    else:
        scales = [1.0]

    for s in scales:
        if s == 1.0:
            new_img = img
        else:
            nh = int(H * s)
            nw = int(W * s)
            new_img = F.interpolate(img, size=(nh, nw), mode='bilinear', align_corners=False)
        nh, nw = new_img.shape[2:]

        with torch.no_grad():
            heatmap, stability, coarse_desc = model.det(new_img)

            # print("nh, nw, heatmap, desc: ", nh, nw, heatmap.shape, coarse_desc.shape)
            if len(heatmap.size()) == 3:
                heatmap = heatmap.unsqueeze(1)
            if len(heatmap.size()) == 2:
                heatmap = heatmap.unsqueeze(0)
                heatmap = heatmap.unsqueeze(1)
            # print(heatmap.shape)
            if heatmap.size(2) != nh or heatmap.size(3) != nw:
                heatmap = F.interpolate(heatmap, size=[nh, nw], mode='bilinear', align_corners=False)

            if stability is not None:
                heatmap = heatmap * stability
                # print('using stability....')

            conf_thresh = conf_th
            nms_dist = 4
            border_remove = 4
            # heatmap = heatmap.data.cpu().numpy().squeeze()
            # xs, ys = np.where(heatmap >= conf_thresh)  # Confidence threshold.
            #
            # pts = np.zeros((3, len(xs)))  # Populate point data sized 3xN.
            # pts[0, :] = ys
            # pts[1, :] = xs
            # pts[2, :] = heatmap[xs, ys]
            #
            # pts, _ = nms_fast(pts, nh, nw, dist_thresh=nms_dist)  # Apply NMS.

            scores = simple_nms(heatmap, nms_radius=nms_dist)
            keypoints = [
                torch.nonzero(s > conf_thresh)
                for s in scores]
            scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]
            # print(keypoints[0].shape)
            keypoints = [torch.flip(k, [1]).float() for k in keypoints]
            scores = scores[0].data.cpu().numpy().squeeze()
            keypoints = keypoints[0].data.cpu().numpy().squeeze()
            pts = keypoints.transpose()
            pts[2, :] = scores
            # print(pts.shape, keypoints.shape)
            # pts = np.zeros((3, keypoints.shape[0]))  # Populate point data sized 3xN.
            # print(len(keypoints), keypoints[0].shape, keypoints[0][0])
            # exit(0)
            # pts[0, :] = keypoints[:, 1]
            # pts[1, :] = keypoints[:, 0]
            # pts[2, :] = scores

            inds = np.argsort(pts[2, :])
            pts = pts[:, inds[::-1]]  # Sort by confidence.
            # Remove points along border.
            bord = border_remove
            toremoveW = np.logical_or(pts[0, :] < bord, pts[0, :] >= (W - bord))
            toremoveH = np.logical_or(pts[1, :] < bord, pts[1, :] >= (H - bord))
            toremove = np.logical_or(toremoveW, toremoveH)
            pts = pts[:, ~toremove]

            # valid_idex = heatmap > conf_thresh
            # valid_score = heatmap[valid_idex]
            # """
            # --- Process descriptor.
            # coarse_desc = coarse_desc.data.cpu().numpy().squeeze()
            D = coarse_desc.size(1)
            if pts.shape[1] == 0:
                desc = np.zeros((D, 0))
            else:
                if coarse_desc.size(2) == nh and coarse_desc.size(3) == nw:
                    desc = coarse_desc[:, :, pts[1, :], pts[0, :]]
                    desc = desc.data.cpu().numpy().reshape(D, -1)
                else:
                    # Interpolate into descriptor map using 2D point locations.
                    samp_pts = torch.from_numpy(pts[:2, :].copy()).to(coarse_desc.device)
                    samp_pts[0, :] = (samp_pts[0, :] / (float(nw) / 2.)) - 1.
                    samp_pts[1, :] = (samp_pts[1, :] / (float(nh) / 2.)) - 1.
                    samp_pts = samp_pts.transpose(0, 1).contiguous()
                    samp_pts = samp_pts.view(1, 1, -1, 2)
                    samp_pts = samp_pts.float()
                    desc = torch.nn.functional.grid_sample(coarse_desc, samp_pts)
                    desc = desc.data.cpu().numpy().reshape(D, -1)
                    desc /= np.linalg.norm(desc, axis=0)[np.newaxis, :]

            if pts.shape[1] == 0:
                continue

            # print(pts.shape, heatmap.shape, new_img.shape, img.shape, nw, nh, W, H)
            pts[0, :] = pts[0, :] * W / nw
            pts[1, :] = pts[1, :] * H / nh
            all_pts.append(np.transpose(pts, [1, 0]))
            all_descs.append(np.transpose(desc, [1, 0]))

    all_pts = np.vstack(all_pts)
    all_descs = np.vstack(all_descs)

    if all_pts.shape[0] == 0:
        return None, None, None

    # keypoints = np.transpose(pts, [1, 0])
    # descriptors = np.transpose(desc, [1, 0])
    # scores = keypoints[:, 2]
    # keypoints = keypoints[:, 0:2]

    keypoints = all_pts[:, 0:2]
    scores = all_pts[:, 2]
    descriptors = all_descs

    # if mask is not None:
    #     id_img = np.int32(mask[:, :, 2]) * 256 * 256 + np.int32(mask[:, :, 1]) * 256 + np.int32(mask[:, :, 0])
    #     labels = id_img[int(keypoints[:, 0])]
    #
    if mask is not None:
        # cv2.imshow("mask", mask)
        # cv2.waitKey(0)
        labels = []
        others = []
        keypoints_with_labels = []
        scores_with_labels = []
        descriptors_with_labels = []
        keypoints_without_labels = []
        scores_without_labels = []
        descriptors_without_labels = []

        id_img = np.int32(mask[:, :, 2]) * 256 * 256 + np.int32(mask[:, :, 1]) * 256 + np.int32(mask[:, :, 0])
        # print(img.shape, id_img.shape)

        for i in range(keypoints.shape[0]):
            x = keypoints[i, 0]
            y = keypoints[i, 1]
            # print("x-y", x, y, int(x), int(y))
            gid = id_img[int(y), int(x)]
            if gid == 0:
                keypoints_without_labels.append(keypoints[i])
                scores_without_labels.append(scores[i])
                descriptors_without_labels.append(descriptors[i])
                others.append(0)
            else:
                keypoints_with_labels.append(keypoints[i])
                scores_with_labels.append(scores[i])
                descriptors_with_labels.append(descriptors[i])
                labels.append(gid)

        # keypoints_with_labels = np.array(keypoints_with_labels)
        # scores_with_labels = np.array(scores_with_labels)
        # descriptors_with_labels = np.array(descriptors_with_labels)
        # labels = np.array(labels, np.int32)
        #
        # keypoints_without_labels = np.array(keypoints_without_labels)
        # scores_without_labels = np.array(scores_without_labels)
        # descriptors_without_labels = np.array(descriptors_without_labels)
        # others = np.array(others, np.int32)

        if topK > 0:
            if topK <= len(keypoints_with_labels):
                idxes = np.array(scores_with_labels, np.float).argsort()[::-1][:topK]
                keypoints = np.array(keypoints_with_labels, np.float)[idxes]
                scores = np.array(scores_with_labels, np.float)[idxes]
                labels = np.array(labels, np.int32)[idxes]
                descriptors = np.array(descriptors_with_labels, np.float)[idxes]
            elif topK >= len(keypoints_with_labels) + len(keypoints_without_labels):
                # keypoints = np.vstack([keypoints_with_labels, keypoints_without_labels])
                # scores = np.vstack([scorescc_with_labels, scores_without_labels])
                # descriptors = np.vstack([descriptors_with_labels, descriptors_without_labels])
                # labels = np.vstack([labels, others])
                keypoints = keypoints_with_labels
                scores = scores_with_labels
                descriptors = descriptors_with_labels
                for i in range(len(others)):
                    keypoints.append(keypoints_without_labels[i])
                    scores.append(scores_without_labels[i])
                    descriptors.append(descriptors_without_labels[i])
                    labels.append(others[i])
            else:
                n = topK - len(keypoints_with_labels)
                idxes = np.array(scores_without_labels, np.float).argsort()[::-1][:n]
                keypoints = keypoints_with_labels
                scores = scores_with_labels
                descriptors = descriptors_with_labels
                for i in idxes:
                    keypoints.append(keypoints_without_labels[i])
                    scores.append(scores_without_labels[i])
                    descriptors.append(descriptors_without_labels[i])
                    labels.append(others[i])
        keypoints = np.array(keypoints, np.float)
        descriptors = np.array(descriptors, np.float)
        # print(keypoints.shape, descriptors.shape)
        return {"keypoints": np.array(keypoints, np.float),
                "descriptors": np.array(descriptors, np.float),
                "scores": np.array(scores, np.float),
                "labels": np.array(labels, np.int32),
                }
    else:
        # print(topK)
        if topK > 0:
            idxes = np.array(scores, dtype=float).argsort()[::-1][:topK]
            keypoints = np.array(keypoints[idxes], dtype=float)
            scores = np.array(scores[idxes], dtype=float)
            descriptors = np.array(descriptors[idxes], dtype=float)

        keypoints = np.array(keypoints, dtype=float)
        scores = np.array(scores, dtype=float)
        descriptors = np.array(descriptors, dtype=float)

        # print(keypoints.shape, descriptors.shape)

        return {"keypoints": np.array(keypoints, dtype=float),
                "descriptors": descriptors,
                "scores": scores,
                }
    # return all_pts, all_descs, all_scores



import os
from .detectors import load_DaD as load_DaD
from .detectors import dedode_detector_S as dedode_detector_S
from .detectors import dedode_detector_B as dedode_detector_B
from .detectors import dedode_detector_L as dedode_detector_L
from .detectors import load_DaDDark as load_DaDDark
from .detectors import load_DaDLight as load_DaDLight
from .types import Detector as Detector
from .types import Matcher as Matcher
from .types import Benchmark as Benchmark

DEBUG_MODE = bool(os.environ.get("DEBUG", False))
RANK = 0
GLOBAL_STEP = 0



import torch
from e2cnn.nn import *
import torch.nn.functional as F

def normalize_coordinates(coords, h, w):
    h=h-1
    w=w-1
    coords=coords.clone().detach()
    coords[:, :, 0]-= w / 2
    coords[:, :, 1]-= h / 2
    coords[:, :, 0]/= w / 2
    coords[:, :, 1]/= h / 2
    return coords

def interpolate_feats(img,pts,feats):
    """
    img : B, 3, H, W
    pts : B, N, 2
    feats : B, C, H', W'
    """
    # compute location on the feature map (due to pooling)
    _, _, h, w = feats.shape
    pool_num = img.shape[-1] / feats.shape[-1]
    pts_warp= pts / pool_num
    pts_norm=normalize_coordinates(pts_warp,h,w)
    pts_norm=torch.unsqueeze(pts_norm, 1)  # b,1,n,2

    # interpolation
    pfeats=F.grid_sample(feats, pts_norm, 'bilinear', align_corners=True)[:, :, 0, :]  # b,f,n
    pfeats=pfeats.permute(0,2,1) # b,n,f

    return pfeats

class RELF_model(torch.nn.Module):
    def __init__(self, args):
        super(RELF_model, self).__init__()
        if args.model == 're_resnet':
            from .reresnet_model.model import ReResNet_wrapper
            self.model = ReResNet_wrapper(args.num_group, depth=18, channels=64)
        else:
            from .rewrn_model.model import E2CNN_wrapper         
            args.num_group = 8   
            self.model = E2CNN_wrapper(args)

        self.in_type = self.model.get_in_type()
        self.max_size_ap = 1024  ## 'max size of adaptive image pyramid.'
        
        self.mode = 'test'

    def forward(self, image, kpts):
        """ multi-scale inference of image pyramid 
            image: [B, 3, H, W], kpts: [B, K, 2]
        """
        images, scales = self.get_scale_space_images(image)

        pfeats_ms = []        
        for sc, image_resize in zip(scales, images):
            kpts_resize = sc * kpts 
            pfeats = self.forward_single_scale(image_resize, kpts_resize)
            pfeats_ms.append(pfeats)
            
        ## We collapse the scale-equivariant features using simple max operation. (scale pooling)
        pfeats_ms = torch.max(torch.stack(pfeats_ms), dim=0)[0]

        return pfeats_ms

    def forward_single_scale(self, image, kpts):
        image = GeometricTensor(image, self.in_type)
        feats = self.model(image)
        pfeats = interpolate_feats(image, kpts, feats)
        return pfeats
    
    def get_scale_space_images(self, image):
        if self.mode == 'train':
            images, scales = ([image], [1.0])
        else:
            images, scales = adaptive_image_pyramid(image, max_size=self.max_size_ap)
        # images, scales = ([image], [1.0])

        return images, scales

    def get_in_type(self):
        return self.in_type

    def switch_mode(self, mode='test'):
        self.mode = mode

## adaptive image pyramid
def adaptive_image_pyramid(img, min_scale=0.0, max_scale=1, min_size=256, max_size=1536, scale_f=2**0.25, verbose=False):
    
    B, _, H, W = img.shape

    ## upsample the input to bigger size.
    s = 1.0
    if max(H, W) < max_size:
        s = max_size / max(H, W)
        max_scale = s
        nh, nw = round(H*s), round(W*s)
        # if verbose:  print(f"extracting at highest scale x{s:.02f} = {nw:4d}x{nh:3d}")
        img = F.interpolate(img, (nh,nw), mode='bilinear', align_corners=False)
        
    ## downsample the scale pyramid
    output = []
    scales = []
    while s+0.001 >= max(min_scale, min_size / max(H,W)):
        if s-0.001 <= min(max_scale, max_size / max(H,W)):
            nh, nw = img.shape[2:]

            if verbose: print(f"extracting at scale x{s:.02f} = {nw:4d}x{nh:3d}")
            output.append(img)
            scales.append(s)
        # print(f"passing the loop x{s:.02f} = {nw:4d}x{nh:3d}")        

        s /= scale_f

        # down-scale the image for next iteration
        nh, nw = round(H*s), round(W*s)
        img = F.interpolate(img, (nh,nw), mode='bilinear', align_corners=False)
    
    return output, scales



if __name__ == "__main__":
    import torch
    model = RELF_model()

    image = torch.randn(2,3,400,400)
    kpts = torch.abs(torch.randn(2,300,2))
    kpts = kpts / torch.max(kpts) * 399
    print(kpts.min(), kpts.max())
    
    ret = model(image, kpts)
    print(ret.shape)
    



import torch
import torch.nn.functional as F

def local_correlation(
    feature0,
    feature1,
    local_radius,
    padding_mode="zeros",
    flow = None,
    sample_mode = "bilinear",
):
    r = local_radius
    K = (2*r+1)**2
    B, c, h, w = feature0.size()
    corr = torch.empty((B,K,h,w), device = feature0.device, dtype=feature0.dtype)
    if flow is None:
        # If flow is None, assume feature0 and feature1 are aligned
        coords = torch.meshgrid(
                (
                    torch.linspace(-1 + 1 / h, 1 - 1 / h, h, device=feature0.device),
                    torch.linspace(-1 + 1 / w, 1 - 1 / w, w, device=feature0.device),
                ),
                indexing = 'ij'
                )
        coords = torch.stack((coords[1], coords[0]), dim=-1)[
            None
        ].expand(B, h, w, 2)
    else:
        coords = flow.permute(0,2,3,1) # If using flow, sample around flow target.
    local_window = torch.meshgrid(
                (
                    torch.linspace(-2*local_radius/h, 2*local_radius/h, 2*r+1, device=feature0.device),
                    torch.linspace(-2*local_radius/w, 2*local_radius/w, 2*r+1, device=feature0.device),
                ),
                indexing = 'ij'
                )
    local_window = torch.stack((local_window[1], local_window[0]), dim=-1)[
            None
        ].expand(1, 2*r+1, 2*r+1, 2).reshape(1, (2*r+1)**2, 2)
    for _ in range(B):
        with torch.no_grad():
            local_window_coords = (coords[_,:,:,None]+local_window[:,None,None]).reshape(1,h,w*(2*r+1)**2,2)
            window_feature = F.grid_sample(
                feature1[_:_+1], local_window_coords, padding_mode=padding_mode, align_corners=False, mode = sample_mode, #
            )
            window_feature = window_feature.reshape(c,h,w,(2*r+1)**2)
        corr[_] = (feature0[_,...,None]/(c**.5)*window_feature).sum(dim=0).permute(2,0,1)
    return corr
